_OP_ALIASES: list[list[str]] = [
    ["__add__", "add"],
    ["__sub__", "sub"],
    ["__mul__", "mult"],
    ["__truediv__", "div"],
    ["__neg__", "negate"],
    ["__pow__", "powBinary"],
    ["__mod__", "modulo"],
    ["__round__", "round"],
    ["__getitem__", "pick", "index"],
    ["__contains__", "contains"],
    ["__len__", "count"],
    ["__ge__", "greaterEqual"],
    ["__gt__", "greater"],
    ["__le__", "lessEqual"],
    ["__lt__", "less"],
    ["__eq__", "equal"],
    ["__ne__", "notEqual"],
    ["startswith", "startsWith"],
    ["endswith", "endsWith"],
    ["isalpha", "isAlpha"],
    ["isalnum", "isAlnum"],
    ["isnumeric", "isNumeric"],
    ["lstrip", "lStrip"],
    ["rstrip", "rStrip"],
    ["in_", "in"],
    # This is Run.set_state. TODO: just fix it.
    ["set_state", "setstate"],
    ["set_output", "setoutput"],
    ["print_", "print"],
    ["await_final_output", "await"],
]

_OP_ALIASES_LOOKUP: dict[str, list[str]] = {}
for alias_set in _OP_ALIASES:
    for op in alias_set:
        _OP_ALIASES_LOOKUP[op] = alias_set


def get_op_aliases(op_name: str) -> list[str]:
    return _OP_ALIASES_LOOKUP.get(op_name, [op_name])



import dataclasses
import typing
from urllib import parse

from . import errors

if typing.TYPE_CHECKING:
    from . import ref_base


@dataclasses.dataclass
class WeaveURI:
    # classvar to identify the URI type
    SCHEME: typing.ClassVar[typing.Optional[str]] = None
    name: str
    version: typing.Optional[str]

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        raise NotImplementedError

    # Used to parse an object URI into it's appropriate WeaveURI subclass
    @classmethod
    def parse(cls: typing.Type["WeaveURI"], uri: str) -> "WeaveURI":
        scheme, netloc, path, params, query_s, fragment = parse.urlparse(uri)
        query = parse.parse_qs(query_s)

        for candidate_class in [cls] + cls.__subclasses__():
            if candidate_class.SCHEME == scheme:
                return candidate_class.from_parsed_uri(
                    uri, scheme, netloc, path, params, query, fragment
                )
        raise errors.WeaveInternalError("invalid scheme ", uri)

    def __str__(self) -> str:
        raise NotImplementedError

    def to_ref(self) -> "ref_base.Ref":
        raise NotImplementedError


# Used when the Weave object is constructed at runtime (eg. weave-builtins or user-defined objects)
@dataclasses.dataclass
class WeaveRuntimeURI(WeaveURI):
    SCHEME = ""

    @classmethod
    def from_parsed_uri(
        cls,
        uri: str,
        schema: str,
        netloc: str,
        path: str,
        params: str,
        query: dict[str, list[str]],
        fragment: str,
    ):
        parts = path.split(":", 1)
        name = parts[0]
        version: typing.Optional[str] = None
        if len(parts) == 2:
            version = parts[1]
        return cls(name, version)

    def __str__(self):
        if self.version is not None:
            return f"{self.name}:{self.version}"
        return self.name



import typing
import dataclasses
import transformers

import weave

from . import hfmodel


# We have to forward-declare the Weave types to avoid circular reference
# issues that weave.type() can't resolve yet.


class HFModelTextGenerationType(hfmodel.HFModelType):
    pass


class FullTextGenerationPipelineOutputType(hfmodel.FullPipelineOutputType):
    def property_types(self):
        return {
            "_model": HFModelTextGenerationType(),
            "model_input": weave.types.String(),
            "model_output": weave.types.List(
                weave.types.TypedDict(
                    {
                        "generated_text": weave.types.String(),
                    }
                )
            ),
        }


class TextGenerationPipelineOutput(typing.TypedDict):
    generated_text: str


@weave.weave_class(weave_type=FullTextGenerationPipelineOutputType)
@dataclasses.dataclass
class FullTextGenerationPipelineOutput(hfmodel.FullPipelineOutput):
    _model: "HFModelTextGeneration"
    model_input: str
    model_output: list[TextGenerationPipelineOutput]


FullTextGenerationPipelineOutputType.instance_classes = FullTextGenerationPipelineOutput


@weave.type()
class FullTextGenerationPanel(weave.Panel):
    id = "FullTextGenerationPanel"
    input_node: weave.Node[FullTextGenerationPipelineOutput]

    @weave.op()
    def render(self) -> weave.panels.Group:
        output = typing.cast(FullTextGenerationPipelineOutput, self.input_node)
        return weave.panels.Group(
            preferHorizontal=True,
            items={
                "input": weave.panels.LabeledItem(
                    label="input", item=output.model_input
                ),
                "output": weave.panels.LabeledItem(
                    label="output", item=output.model_output
                ),
            },
        )


@weave.weave_class(weave_type=HFModelTextGenerationType)
@dataclasses.dataclass
class HFModelTextGeneration(hfmodel.HFModel):
    @weave.op()
    def pipeline(
        self,
    ) -> transformers.pipelines.Pipeline:
        return transformers.pipeline(
            self._pipeline_tag,
            model=self._id,
        )

    @weave.op()
    # TODO: There are many more arguments for text generation. Some of the arguments
    #    are pipeline arguments, others are passed to the underlying model's .generate()
    #    method. Need to find the docs for each of those.
    #    It'd be nice to call these different types, and therefore have different signatures
    #    for this call!
    def call(self, input: str) -> FullTextGenerationPipelineOutput:
        output = weave.use(self.pipeline())(input)
        return FullTextGenerationPipelineOutput(self, input, output)

    @weave.op()
    # TODO: There are many more arguments for text generation. Some of the arguments
    #    are pipeline arguments, others are passed to the underlying model's .generate()
    #    method. Need to find the docs for each of those.
    #    It'd be nice to call these different types, and therefore have different signatures
    #    for this call!
    def call_list(
        self, input: typing.List[str]
    ) -> typing.List[FullTextGenerationPipelineOutput]:
        output = map(self.pipeline(), input)
        return [
            FullTextGenerationPipelineOutput(self, i, o)
            for (i, o) in zip(input, output)
        ]


HFModelTextGenerationType.instance_classes = HFModelTextGeneration



# We need this boilerplate for all ecosystem packages for now.

from weave import context_state
import logging

logging.getLogger("ecosystem_example").setLevel(logging.ERROR)

loading_builtins_token = context_state.set_loading_built_ins()

try:
    from .ops import *
finally:
    context_state.clear_loading_built_ins(loading_builtins_token)



import gc
import os
import logging
import datetime
import threading
import traceback
import signal
import sys
import typing
import pathlib
from types import FrameType


from . import environment


def dump_folder() -> pathlib.Path:
    raw_folder = os.environ.get("WEAVE_DEBUG_DUMP_FOLDER", "")
    folder = pathlib.Path(raw_folder)
    tmp = pathlib.Path("/tmp")
    if raw_folder == "" or not folder.exists():
        if raw_folder == "":
            logging.warning(
                f"WEAVE_DEBUG_DUMP_FOLDER {folder} does not exist, using /tmp"
            )
        folder = tmp
    return folder.absolute()


def make_output_path(root: str) -> str:
    return str(
        dump_folder()
        / f"{root}-{os.getpid()}-{datetime.datetime.now().isoformat()}.txt".replace(
            " ", "_"
        )
    )


def dump_stack_traces(signal: int, frame: typing.Optional[FrameType]) -> None:
    """Function to be executed when the signal is received."""

    id_to_name = dict([(th.ident, th.name) for th in threading.enumerate()])
    code = []
    for threadId, stack in sys._current_frames().items():
        thread_name = id_to_name.get(threadId, "")
        if thread_name == "MainThread" and frame is not None:
            stack = frame
        code.append("\n# Thread: %s(%d)" % (thread_name, threadId))
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                code.append("  %s" % (line.strip()))

    output_path = make_output_path("weave_server_stack_dump")
    with open(output_path, "w+") as f:
        f.write("\n".join(code))


def objgraph_getnewids(signal: int, frame: typing.Optional[FrameType]) -> None:
    import objgraph

    gc.collect()
    output_path = make_output_path("weave_server_objgraph_newids")
    with open(output_path, "w") as f:
        objgraph.get_new_ids(limit=100, file=f)


def objgraph_showgrowth(signal: int, frame: typing.Optional[FrameType]) -> None:
    import objgraph

    gc.collect()
    output_path = make_output_path("weave_server_objgraph_growth")
    with open(output_path, "w") as f:
        objgraph.show_growth(limit=100, file=f)


def install_signal_handler_to_be_called_after_existing(
    signum: int, handler: typing.Callable[[int, typing.Optional[FrameType]], None]
) -> None:
    cur_sig_handler = signal.getsignal(signum)
    if (
        cur_sig_handler is not None
        and cur_sig_handler != signal.SIG_DFL
        and cur_sig_handler != signal.SIG_IGN
    ):

        def new_handler(signal: int, frame: typing.Optional[FrameType]) -> None:
            cur_sig_handler(signal, frame)  # type: ignore
            handler(signal, frame)

        signal.signal(signum, new_handler)
    else:
        signal.signal(signum, handler)


def install_signal_handlers() -> None:
    if (
        environment.memdump_sighandler_enabled()
        or environment.stack_dump_sighandler_enabled()
    ):
        # Here we set the SIGUSR1 signal handler to our function dump_stack_traces
        # To use, send a SIGUSR1 signal to set a baseline, then do some requests.
        # Then send a SIGUSR2 signal to drop the server into pdb and do
        # import objgraph
        # obj_ids = objgraph.get_new_ids()
        # This will contain all the ids of objects that have been created since
        # the last call to objgraph.get_new_ids()
        # Then you can inspect objects like:
        # obj_id = obj_ids['TypedDict'][0]
        # obj = objgraph.at(obj_id)
        # objgraph.show_backrefs([obj], max_depth=15)
        #
        # Other useful objgraph commands:
        # objgraph.show_most_common_types(limit=20)
        # obj = objgraph.by_type('TypedDict')[100]

        def sigusr1_handler(signal: int, frame: typing.Optional[FrameType]) -> None:
            if environment.memdump_sighandler_enabled():
                objgraph_getnewids(signal, frame)
            if environment.stack_dump_sighandler_enabled():
                dump_stack_traces(signal, frame)

        install_signal_handler_to_be_called_after_existing(
            signal.SIGUSR1, sigusr1_handler
        )

        if environment.memdump_sighandler_enabled():
            install_signal_handler_to_be_called_after_existing(
                signal.SIGUSR2, objgraph_showgrowth
            )



"""
This file contains the Weave definition for KerasModel, allowing users to
save & publish Keras models. This handles serializations as well as extracting
the type definition for the model (input and output tensors). This is a work
in progress.

TODOS:
    - [ ] Add a single `call` method to the model class where the input data type is
        dependent on the model type. Shawn has ideas around how to do this (to make the
        type of arguments dependent on previous arguments). Moreover, the return type
        will also be dependent on the model. Currently we have a nasty `call_string_to_number` and `call_string_to_string` 
        as placeholders.
    - [ ] Figure out the correct way to do batching. There are two considerations:
        1. We allow tensors to have any number of `None` dimensions. We may want to remove this allowance, or restrict to just a single `None` dimension.
        2. We should make the `call*` functions mappable and if the modle has a batch dimension, bulk call on the model. - right now we just hard-code single batches.
    - [ ] The `image_classification` method is purely an example. It has some issues:
        1. The pre- and post- processing is hard coded for the example model - we need a way to properly handle a DAG/Series of transforms.
        2. The input is a URL - we need a way to nicely interop with the rest of the Images in our ecosystem.
"""


from dataclasses import dataclass
from enum import Enum
from typing import Optional, Type, Union
import typing
from tensorflow import keras
from keras.engine import keras_tensor
from tensorflow.python.framework import dtypes

import weave


class DTYPE_NAME(Enum):
    NUMBER = "number"
    BOOL = "bool"
    STRING = "string"
    UNMAPPED = "unmapped"


# see https://www.tensorflow.org/api_docs/python/tf/dtypes
DTYPE_TO_TYPE: dict[dtypes.bfloat16, weave.types.Type] = {
    dtypes.bfloat16: weave.types.Number(),
    dtypes.double: weave.types.Number(),
    dtypes.float16: weave.types.Number(),
    dtypes.float32: weave.types.Number(),
    dtypes.float64: weave.types.Number(),
    dtypes.half: weave.types.Number(),
    dtypes.int16: weave.types.Number(),
    dtypes.int32: weave.types.Number(),
    dtypes.int64: weave.types.Number(),
    dtypes.int8: weave.types.Number(),
    dtypes.qint16: weave.types.Number(),
    dtypes.qint32: weave.types.Number(),
    dtypes.qint8: weave.types.Number(),
    dtypes.quint16: weave.types.Number(),
    dtypes.quint8: weave.types.Number(),
    dtypes.uint16: weave.types.Number(),
    dtypes.uint32: weave.types.Number(),
    dtypes.uint64: weave.types.Number(),
    dtypes.uint8: weave.types.Number(),
    dtypes.bool: weave.types.Boolean(),
    dtypes.string: weave.types.String(),
    dtypes.complex128: weave.types.UnknownType(),
    dtypes.complex64: weave.types.UnknownType(),
    dtypes.resource: weave.types.UnknownType(),
    dtypes.variant: weave.types.UnknownType(),
}


def shape_to_dict(shape: typing.List[typing.Optional[int]]) -> weave.types.TypedDict:
    return weave.types.TypedDict(
        {
            f"{shape_ndx}": (
                weave.types.NoneType()
                if dim is None
                else weave.types.Const(weave.types.Number(), dim)
            )
            for shape_ndx, dim in enumerate(shape)
        }
    )


def shape_to_list(
    shape: typing.List[typing.Optional[int]], inner_type: weave.types.Type
) -> weave.types.Type:
    if len(shape) == 0:
        return inner_type
    else:
        # Once List supports length params, add it here
        return weave.types.List(shape_to_list(shape[1:], inner_type))


@dataclass(frozen=True)
class KerasTensorType(weave.types.Type):
    instance_classes = keras_tensor.KerasTensor
    instance_class = keras_tensor.KerasTensor

    shape: weave.types.Type = weave.types.Any()
    data_type: weave.types.Type = weave.types.Any()
    # Temporary type field to hold the "vector" like type
    weave_vector_type: weave.types.Type = weave.types.Any()

    @classmethod
    def type_of_instance(cls, obj):
        return cls(
            shape=shape_to_dict(obj.shape.as_list()),
            data_type=DTYPE_TO_TYPE[obj.dtype],
            weave_vector_type=shape_to_list(
                obj.shape.as_list(), DTYPE_TO_TYPE[obj.dtype]
            ),
        )

    # This is just a helper function for building the type of a KerasTensor
    @classmethod
    def from_list(
        cls: Type["KerasTensorType"],
        shape: list[Union[None, int]],
        data_type: weave.types.Type = weave.types.Any(),
    ) -> "KerasTensorType":
        return cls(
            shape=shape_to_dict(shape),
            data_type=data_type,
            weave_vector_type=shape_to_list(shape, data_type),
        )


@dataclass(frozen=True)
class KerasModel(weave.types.Type):
    instance_classes = keras.Model
    instance_class = keras.Model

    inputs_type: weave.types.Type = weave.types.Any()  # list[KerasTensorType]
    outputs_type: weave.types.Type = weave.types.Any()  # list[KerasTensorType]

    @classmethod
    def type_of_instance(cls, obj):
        inputs_as_dict = {
            f"{k}": weave.types.TypeRegistry.type_of(v)
            for k, v in enumerate(obj.inputs)
        }
        outputs_as_dict = {
            f"{k}": weave.types.TypeRegistry.type_of(v)
            for k, v in enumerate(obj.outputs)
        }
        return cls(
            weave.types.TypedDict(inputs_as_dict),
            weave.types.TypedDict(outputs_as_dict),
        )

    def save_instance(self, obj, artifact, name):
        with artifact.new_dir(f"{name}") as dirpath:
            obj.save(dirpath)

    def load_instance(self, artifact, name, extra=None):
        return keras.models.load_model(artifact.path(name))

    # This is just a helper function for building the type of a KerasModel
    @classmethod
    def make_type(
        cls: Type["KerasModel"],
        inputs_def: Optional[
            list[typing.Union[KerasTensorType, weave.types.Any]]
        ] = None,
        outputs_def: Optional[
            list[typing.Union[KerasTensorType, weave.types.Type]]
        ] = None,
    ) -> "KerasModel":
        inputs = (
            weave.types.TypedDict(
                {f"{input_ndx}": t for input_ndx, t in enumerate(inputs_def)}
            )
            if inputs_def is not None
            else weave.types.Any()
        )
        outputs = (
            weave.types.TypedDict(
                {f"{input_ndx}": t for input_ndx, t in enumerate(outputs_def)}
            )
            if outputs_def is not None
            else weave.types.Any()
        )
        return cls(inputs, outputs)


def byte_vector_to_string(maybe_byte_vector: typing.Any) -> typing.Union[list, str]:
    if isinstance(maybe_byte_vector, bytes):
        return maybe_byte_vector.decode("utf-8")
    elif isinstance(maybe_byte_vector, list):
        return [byte_vector_to_string(item) for item in maybe_byte_vector]
    else:
        return maybe_byte_vector


# Remaining limitations:
# - Batching is hard coded
# - Input type is hard coded
# - Single output and single input layer is hard coded
# - Vectors are not sized
@weave.op(
    input_type={
        "model": KerasModel.make_type(
            [KerasTensorType.from_list([None, 1], weave.types.String())],
            [weave.types.Any()],
        ),
        "input": weave.types.String(),
    },
    output_type=lambda input_types: input_types["model"]
    .outputs_type.property_types["0"]
    .weave_vector_type.object_type,
)
def call_string(model, input):
    res = model.predict([[input]]).tolist()[0]
    # Special case for strings: we need to convert the bytes to a string
    return byte_vector_to_string(res)


# Added back by Shawn to make notebook work for now.
# call_string has a callable output type. WeaveJS doesn't seem to be able
# to make use of it.
# TODO: remove
@weave.op(
    input_type={
        "model": KerasModel.make_type(
            [KerasTensorType.from_list([None, 1], weave.types.String())],
            [weave.types.String()],
        ),
        "input": weave.types.String(),
    },
)
def call_string_to_number(model, input) -> int:
    return model.predict([[input]]).tolist()[0][0]



import typing
from weave import box, graph
from weave.artifact_wandb import likely_commit_hash
from weave.language_features.tagging import tag_store
from weave.node_ref import ref_to_node
from weave.uris import WeaveURI
from . import mappers
from . import ref_base
from . import weave_types as types
from . import errors
from . import mappers_python_def
from .language_features.tagging import tagged_value_type
import dataclasses
from weave import weave_internal, context, storage
from .ops_primitives import weave_api


class RefToPyRef(mappers.Mapper):
    def apply(self, obj: ref_base.Ref):
        if _uri_is_local_artifact(obj.uri):
            obj = _local_ref_to_published_ref(obj)

        return obj


class FunctionToPyFunction(mappers.Mapper):
    def apply(self, obj):
        res = graph.map_nodes_full([obj], _node_publish_mapper)[0]
        return res


class ObjectToPyDict(mappers_python_def.ObjectToPyDict):
    def apply(self, obj):
        try:
            res = super().apply(obj)
            copy_obj = dataclasses.copy.copy(obj)
            for prop_name, prop_serializer in self._property_serializers.items():
                if prop_serializer is not None:
                    setattr(copy_obj, prop_name, res[prop_name])
            obj = copy_obj
        except Exception as e:
            print(e)
            pass
        return obj


class UnionToPyUnion(mappers_python_def.UnionToPyUnion):
    def apply(self, obj):
        res = super().apply(obj)
        if not isinstance(res, dict):
            raise errors.WeaveSerializeError("")
        if isinstance(obj, dict):
            return {k: v for k, v in res.items() if k != "_union_id"}
        else:
            return res["_val"]


class TaggedValueToPy(tagged_value_type.TaggedValueToPy):
    def apply(self, obj: typing.Any) -> dict:
        res = super().apply(obj)
        value = self._value_serializer.apply(res["_value"])
        tags = self._tag_serializer.apply(res["_tag"])
        value = box.box(value)
        tag_store.add_tags(value, tags)
        return value


def map_to_python_remote_(type, mapper, artifact, path=[], mapper_options=None):
    if isinstance(type, types.Function):
        return FunctionToPyFunction(type, mapper, artifact, path)
    elif isinstance(type, types.RefType):
        return RefToPyRef(type, mapper, artifact, path)

    elif isinstance(type, types.TypedDict):
        return mappers_python_def.TypedDictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.Dict):
        return mappers_python_def.DictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.List):
        return mappers_python_def.ListToPyList(type, mapper, artifact, path)
    elif isinstance(type, types.UnionType):
        return UnionToPyUnion(type, mapper, artifact, path)
    elif isinstance(type, types.ObjectType):
        return ObjectToPyDict(type, mapper, artifact, path)
    elif isinstance(type, tagged_value_type.TaggedValueType):
        return TaggedValueToPy(type, mapper, artifact, path)
    elif isinstance(type, types.Const):
        return mappers_python_def.ConstToPyConst(type, mapper, artifact, path)

    return mappers.Mapper(type, mapper, artifact, path)


map_to_python_remote = mappers.make_mapper(map_to_python_remote_)


def _node_publish_mapper(node: graph.Node) -> typing.Optional[graph.Node]:
    if _node_is_op_get(node):
        node = typing.cast(graph.OutputNode, node)
        uri = _uri_of_get_node(node)
        if uri is not None and _uri_is_local_artifact(uri):
            # Be sure to merge the node if needed before continuing with the publish.
            if weave_api.get_merge_spec_uri(uri):
                return _node_publish_mapper(weave_api.get(weave_api._merge(uri)))
            return _local_op_get_to_published_op_get(node)
    return node


def _node_is_op_get(node: graph.Node) -> bool:
    return isinstance(node, graph.OutputNode) and node.from_op.name == "get"


def _uri_of_get_node(node: graph.OutputNode) -> typing.Optional[str]:
    uri_node = node.from_op.inputs.get("uri")
    if isinstance(uri_node, graph.ConstNode) and isinstance(uri_node.val, str):
        return uri_node.val
    return None


def _uri_is_local_artifact(uri: str) -> bool:
    return uri.startswith("local-artifact://")


def _name_and_branch_from_node(
    node: typing.Optional[graph.Node],
) -> typing.Tuple[typing.Optional[str], typing.Optional[str]]:
    name: typing.Optional[str] = None
    version: typing.Optional[str] = None
    if node is not None and _node_is_op_get(node):
        uri_str = _uri_of_get_node(node)  # type: ignore
        if uri_str is not None:
            uri = WeaveURI.parse(uri_str)
            name = uri.name
            if uri.version is not None and not likely_commit_hash(uri.version):
                version = uri.version

    return (name, version)


def _local_op_get_to_pub_ref(node: graph.Node) -> ref_base.Ref:
    obj = weave_internal.use(node, context.get_client())
    name, version = _name_and_branch_from_node(node)

    return storage._direct_publish(
        obj, name, branch_name=version, assume_weave_type=node.type
    )


def _local_op_get_to_published_op_get(node: graph.Node) -> graph.Node:
    pub_ref = _local_op_get_to_pub_ref(node)
    new_node = ref_to_node(pub_ref)

    if new_node is None:
        raise errors.WeaveSerializeError(
            f"Failed to serialize {node} to published node"
        )

    return new_node


def _local_ref_to_published_ref(ref: ref_base.Ref) -> ref_base.Ref:
    node = ref_to_node(ref)

    if node is None:
        raise errors.WeaveSerializeError(f"Failed to serialize {ref} to published ref")
    return _local_op_get_to_pub_ref(node)



import threading
import multiprocessing

import queue
import numpy as np
from .. import errors
from .. import environment
from .. import context_state
import typing
import logging
import warnings

from sklearn.manifold import TSNE
from sklearn.decomposition import PCA

umap_lib = {}

DEFAULT_TIMEOUT_SEC = environment.projection_timeout_sec()

ResultQueueItemType = typing.Union[Exception, np.ndarray]


def _get_umap():
    # Lazily load a cached version of UMAP - umap import
    # time is quite expensive so we want to do it once and
    # only when needed
    if "lib" not in umap_lib:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import umap

            umap_lib["lib"] = umap
    return umap_lib["lib"]


def perform_2D_projection(
    np_array_of_embeddings: np.ndarray,
    projectionAlgorithm: str,
    algorithmOptions: dict,
) -> np.ndarray:
    if len(np_array_of_embeddings.shape) != 2:
        raise errors.WeaveInternalError(
            f"The input to the 2D projection must be a 2D array of embeddings, found {np_array_of_embeddings.shape}"
        )
    if projectionAlgorithm == "pca":
        projection = perform_2D_projection_pca(
            np_array_of_embeddings, algorithmOptions.get("pca", {})
        )
    elif projectionAlgorithm == "tsne":
        projection = perform_2D_projection_tsne(
            np_array_of_embeddings, algorithmOptions.get("tsne", {})
        )
    elif projectionAlgorithm == "umap":
        projection = perform_2D_projection_umap(
            np_array_of_embeddings, algorithmOptions.get("umap", {})
        )
    else:
        raise Exception("Unknown projection algorithm: " + projectionAlgorithm)
    return projection


def perform_2D_projection_async(
    np_array_of_embeddings: np.ndarray,
    projectionAlgorithm: str,
    algorithmOptions: dict,
    result_queue: queue.Queue[ResultQueueItemType],
):
    try:
        projection = perform_2D_projection(
            np_array_of_embeddings, projectionAlgorithm, algorithmOptions
        )
        result_queue.put(projection)
    except Exception as e:
        result_queue.put(e)


def perform_2D_projection_with_timeout(
    np_array_of_embeddings: np.ndarray,
    projectionAlgorithm: str,
    algorithmOptions: dict,
    timeout: typing.Optional[typing.Union[int, float]] = DEFAULT_TIMEOUT_SEC,
) -> np.ndarray:
    if timeout is None:
        return perform_2D_projection(
            np_array_of_embeddings, projectionAlgorithm, algorithmOptions
        )

    # otherwise run it in another process and kill it if it goes over time

    result_queue: queue.Queue[ResultQueueItemType] = multiprocessing.Queue()
    target = multiprocessing.Process(
        target=perform_2D_projection_async,
        args=(
            np_array_of_embeddings,
            projectionAlgorithm,
            algorithmOptions,
            result_queue,
        ),
        name="projection",
    )
    target.start()

    try:
        result = result_queue.get(timeout=timeout)
    except queue.Empty:
        target.kill()
        logging.error(
            f"Projection timed out after {timeout} seconds, killing process {target.pid}, returning empty projection",
        )

        return np.zeros((len(np_array_of_embeddings), 2))
    if isinstance(result, Exception):
        raise result

    return result


def limit_embedding_dimensions(
    np_array_of_embeddings: np.ndarray, max_dimensions: int = 50
) -> np.ndarray:
    max_dimensions = min(max_dimensions, len(np_array_of_embeddings))
    if np_array_of_embeddings.shape[1] > max_dimensions:
        return PCA(n_components=max_dimensions).fit_transform(np_array_of_embeddings)
    return np_array_of_embeddings


def perform_2D_projection_pca(
    np_array_of_embeddings: np.ndarray, options: dict
) -> np.ndarray:
    model = PCA(n_components=2)
    return model.fit_transform(np_array_of_embeddings)


def perform_2D_projection_tsne(
    np_array_of_embeddings: np.ndarray, options
) -> np.ndarray:
    n_samples = len(np_array_of_embeddings)
    return TSNE(
        n_components=2,
        perplexity=min(n_samples - 1, options.get("perplexity", 30)),
        n_iter=min(1000, max(500, options.get("iterations", 1000))),
        learning_rate=options.get("learningRate", "auto"),
        init="random",
    ).fit_transform(limit_embedding_dimensions(np_array_of_embeddings))


# TODO: we need lock stats in datadog
# UMAP is not thread safe, it crashes the server when called at the same
# time on parallel threads.
umap_lock = threading.Lock()


def perform_2D_projection_umap(
    np_array_of_embeddings: np.ndarray, options: dict
) -> np.ndarray:
    model = _get_umap().UMAP(
        n_components=2,
        # Letting the library choose defaults
        # n_neighbors=min(n_samples - 1, options.get("neighbors", 15)),
        # min_dist=options.get("minDist", 0.1),
        # spread=options.get("spread", 1.0),
        init="random",
    )
    with umap_lock:
        return model.fit_transform(np_array_of_embeddings)



import json
import dataclasses
import enum
import pathlib
import warnings
import os
import pathlib
import logging
import logging.config
from logging.handlers import WatchedFileHandler
from flask.logging import wsgi_errors_stream
from pythonjsonlogger import jsonlogger
import typing
import contextlib
import contextvars
import warnings

from . import environment


class LogFormat(enum.Enum):
    PRETTY = "pretty"
    DATADOG = "datadog"


@dataclasses.dataclass
class LogSettings:
    format: LogFormat
    level: typing.Optional[int]


_log_indent: contextvars.ContextVar[int] = contextvars.ContextVar(
    "_log_indent", default=0
)


def increment_indent() -> contextvars.Token[int]:
    indent = _log_indent.get()
    return _log_indent.set(indent + 1)


def reset_indent(token: contextvars.Token[int]) -> None:
    _log_indent.reset(token)


@contextlib.contextmanager
def indent_logs() -> typing.Iterator:
    token = increment_indent()
    try:
        yield
    finally:
        reset_indent(token)


def get_indent() -> int:
    return _log_indent.get()


class IndentFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        indent = get_indent()
        record.indent = "  " * indent  # type: ignore
        return True


default_log_format = "[%(asctime)s] %(levelname)s in %(module)s (Thread Name: %(threadName)s): %(indent)s%(message)s"


def get_logdir() -> typing.Optional[str]:
    logdir = "/tmp/weave/log"
    try:
        pathlib.Path(logdir).mkdir(exist_ok=True, parents=True)
    except OSError:
        warnings.warn(
            f"weave: Unable make log dir '{logdir}'. Filesystem logging will be disabled for "
            f"the remainder of this session. To enable filesystem logging, ensure the path is writable "
            f"and restart the server."
        )
        return None
    return logdir


def get_logfile_path(logfile_path: str) -> typing.Optional[str]:
    logdir = get_logdir()
    if logdir is None:
        return None
    full_path = pathlib.Path(logdir) / logfile_path
    try:
        full_path.touch()
    except OSError:
        warnings.warn(
            f"weave: Unable to touch logfile at '{full_path}'. Filesystem logging will be disabled for "
            f"the remainder of this session. To enable filesystem logging, ensure the path is writable "
            f"and restart the server."
        )
        return None
    return str(full_path)


def default_log_filename() -> typing.Optional[str]:
    pid = os.getpid()
    return get_logfile_path(f"{pid}.log")


def env_log_level() -> int:
    level_str_from_env = os.environ.get("WEAVE_LOG_LEVEL")
    if level_str_from_env:
        level_int_from_env = logging.getLevelName(level_str_from_env)
        if isinstance(level_int_from_env, int):
            return level_int_from_env

    if os.environ.get("WEAVE_SERVER_ENABLE_LOGGING"):
        return logging.INFO

    # Default to only showing ERROR logs
    return logging.ERROR


def silence_mpl() -> None:
    mpl_logger = logging.getLogger("matplotlib")
    if mpl_logger:
        mpl_logger.setLevel(logging.CRITICAL)


def set_global_log_level(level: int) -> None:
    # Unused, but runs through all existing loggers and sets their level to the given level
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    for logger_name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)


def print_handlers(logger: logging.Logger) -> None:
    # For debugging
    print(f"Handlers for logger '{logger.name}':")
    for handler in logger.handlers:
        print(f"  {handler}")


def print_all_handlers() -> None:
    # For debugging
    root_logger = logging.getLogger()
    print_handlers(root_logger)

    for logger_name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(logger_name)
        print_handlers(logger)


class WeaveJSONEncoder(jsonlogger.JsonEncoder):
    def default(self, obj: typing.Any) -> typing.Any:
        if obj is None:
            # This is needed because datadog strips keys with null values from logs
            return "<<_WEAVE_NONE_>>"
        return super().default(obj)  # type: ignore[no-untyped-call]


def setup_handler(handler: logging.Handler, settings: LogSettings) -> None:
    formatter = logging.Formatter(default_log_format)
    if settings.format == LogFormat.DATADOG:
        formatter = jsonlogger.JsonFormatter(
            "%(levelname)s [%(name)s] [%(filename)s:%(lineno)d] "
            "[dd.trace_id=%(dd.trace_id)s dd.span_id=%(dd.span_id)s] %(message)s",
            timestamp=True,
            json_encoder=WeaveJSONEncoder,
        )  # type: ignore[no-untyped-call]
    handler.addFilter(IndentFilter())
    handler.setFormatter(formatter)
    if settings.level is not None:
        handler.setLevel(settings.level)


_LOGGING_CONFIGURED = False


def enable_stream_logging(
    logger: typing.Optional[logging.Logger] = None,
    wsgi_stream_settings: typing.Optional[LogSettings] = None,
    pid_logfile_settings: typing.Optional[LogSettings] = None,
    server_logfile_settings: typing.Optional[LogSettings] = None,
) -> None:
    global _LOGGING_CONFIGURED
    _LOGGING_CONFIGURED = True

    handler: logging.Handler
    logger = logger or logging.getLogger("root")

    if wsgi_stream_settings is not None:
        handler = logging.StreamHandler(wsgi_errors_stream)  # type: ignore
        setup_handler(handler, wsgi_stream_settings)
        logger.addHandler(handler)

    if pid_logfile_settings is not None:
        log_filename = default_log_filename()
        if log_filename:
            handler = WatchedFileHandler(log_filename, mode="w")
            setup_handler(handler, pid_logfile_settings)
            logger.addHandler(handler)

    if server_logfile_settings is not None:
        log_filename = get_logfile_path(f"server.log")
        if log_filename:
            handler = logging.FileHandler(log_filename, mode="w")
            setup_handler(handler, server_logfile_settings)
            logger.addHandler(handler)


def configure_logger() -> None:
    if _LOGGING_CONFIGURED:
        return
    # Disable ddtrace logs, not sure why they're turned on.
    ddtrace_logs = logging.getLogger("ddtrace")
    ddtrace_logs.setLevel(logging.ERROR)

    logger = logging.getLogger("root")
    logger.setLevel(env_log_level())

    # Remove StreamHandler from the root logger to avoid stdout logging
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            logger.removeHandler(handler)

    dd_env = os.getenv("DD_ENV")
    enable_datadog = dd_env
    if not enable_datadog:
        # This is the standard path for users.
        if os.getenv("WEAVE_SERVER_ENABLE_LOGGING"):
            # WEAVE_SERVER_ENABLE_LOGGING forces the logs to go to the wsgi
            # stream which will go to the console. We set this flag in our
            # server start scripts, so that if you run the server in its own
            # terminal, you get the logs.
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(LogFormat.PRETTY, level=None),
                pid_logfile_settings=LogSettings(LogFormat.PRETTY, logging.INFO),
            )
        else:
            # This is the default case. Log to a file for this process, but
            # do not write to standard out/error. This is important because
            # when you run Weave in a notebook, it'll create a server, but
            # we don't want the logs to show up in the notebook.
            enable_stream_logging(
                logger,
                pid_logfile_settings=LogSettings(LogFormat.PRETTY, logging.INFO),
            )
    else:
        if dd_env == "ci":
            # CI expects logs in the pid logfile
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(LogFormat.PRETTY, logging.INFO),
                pid_logfile_settings=LogSettings(LogFormat.PRETTY, logging.INFO),
            )
        elif environment.wandb_production():
            # Only log in the datadog format to the wsgi stream
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(LogFormat.DATADOG, level=None),
            )
        else:
            # Otherwise this is dev mode with datadog logging turned on.
            # Log to standard out and a known filename that the datadog
            # agent can watch.
            enable_stream_logging(
                logger,
                wsgi_stream_settings=LogSettings(LogFormat.PRETTY, level=None),
                server_logfile_settings=LogSettings(LogFormat.DATADOG, level=None),
            )



# In its own file to avoid

import pyarrow as pa

from ..api import op
from .. import weave_types as types

from .list_ import ArrowWeaveList

py_range = range


@op(name="range")
def range(start: int, stop: int, step: int) -> ArrowWeaveList[int]:
    return ArrowWeaveList(pa.array(py_range(start, stop, step)), types.Int())



from . import panel
from .. import graph


class VerticalContainerConfig:
    def __init__(self):
        self._items = []

    def add(self, item):
        self._items.append(item)


class SidebarContainerConfig:
    def __init__(self):
        self.sidebar = VerticalContainerConfig()
        self.content = VerticalContainerConfig()


class SidebarContainer(panel.Panel):
    def __init__(self, input_node=graph.VoidNode(), config=None):
        self.input_node = input_node
        self._config = config
        if self._config is None:
            self._config = SidebarContainerConfig()

    @property
    def config(self):
        return self._config.to_json()

    # This is an op!
    @property
    def value(self):
        return self._config.value



import weave


def test_difference_type():
    diff = weave.ops.difference(
        [weave.const("a"), weave.const("b")], [weave.const("b")]
    )
    # When difference operations on Const Union types, we get a const union
    # type back with the difference applied (ie difference works at type level
    # when types are available as Consts)
    assert diff.type == weave.types.List(
        weave.types.union(weave.types.Const(weave.types.String(), "a"))
    )



import atexit
import contextlib
import datetime
import json
import logging
import os
import queue
import random
import shutil
import threading
import typing
import uuid

from wandb.sdk.lib.paths import LogicalPath
from wandb.sdk.lib.printer import get_printer
from wandb.sdk.lib.ipython import _get_python_type


from .wandb_lite_run import InMemoryLazyLiteRun

from .. import runfiles_wandb
from .. import storage
from .. import weave_types
from .. import artifact_base
from .. import file_util
from .. import graph
from .. import errors
from ..core_types.stream_table_type import StreamTableType
from ..ops_domain import stream_table_ops
from ..ops_primitives import weave_api

if typing.TYPE_CHECKING:
    from wandb.sdk.internal.file_pusher import FilePusher


# Shawn recommended we only encode leafs, but in my testing, nested structures
# are not handled as well in in gorilla and we can do better using just weave.
# Uncomment the below to use gorilla for nested structures.
TRUST_GORILLA_FOR_NESTED_STRUCTURES = True

# Weave types are parametrized, but gorilla expects just simple strings. We could
# send the top-level string over the wire, but this fails to encode type specifics
# and therefore loses information. With this flag, we instead stringify the JSON type
# and send that over the wire. This is a bit of a hack, but it works.
ENCODE_ENTIRE_TYPE = True
TYPE_ENCODE_PREFIX = "_wt_::"

ROW_TYPE = typing.Union[dict, list[dict]]


class WandbLiveRunFiles(runfiles_wandb.WandbRunFiles):
    _file_pusher: typing.Optional["FilePusher"] = None
    _temp_dir: typing.Optional[str] = None

    def set_file_pusher(self, pusher: "FilePusher") -> None:
        # It is the responsibility of the caller to ensure this file pusher is
        # correctly associated with the corresponding run.
        self._file_pusher = pusher

    def temp_dir(self) -> str:
        if self._temp_dir is None:
            rand_part = "".join(random.choice("0123456789ABCDEF") for _ in range(16))
            self._temp_dir = os.path.join(
                runfiles_wandb.wandb_run_dir(), "upload_cache", f"tmp_{rand_part}"
            )
            os.makedirs(self._temp_dir, exist_ok=True)
        return self._temp_dir

    def cleanup(self) -> None:
        if self._temp_dir is not None:
            shutil.rmtree(self._temp_dir)
            self._temp_dir = None

    def __del__(self) -> None:
        self.cleanup()

    @contextlib.contextmanager
    def new_file(
        self, path: str, binary: bool = False
    ) -> typing.Generator[typing.IO, None, None]:
        if self._file_pusher is None:
            raise ValueError(
                "WandbLiveRunFiles must be associated with a file pusher to use new_file"
            )

        dir_path = self.temp_dir()
        file_path = os.path.join(dir_path, path)
        with file_util.safe_open(file_path, "wb" if binary else "w") as file:
            yield file
            self._file_pusher.file_changed(LogicalPath(path), file_path)


class _StreamTableSync:
    _lite_run: InMemoryLazyLiteRun
    _table_name: str
    _project_name: str
    _entity_name: str

    _artifact: WandbLiveRunFiles

    _weave_stream_table: StreamTableType
    _weave_stream_table_ref: artifact_base.ArtifactRef

    _client_id: str

    def __init__(
        self,
        table_name: str,
        *,
        project_name: typing.Optional[str] = None,
        entity_name: typing.Optional[str] = None,
        _disable_async_file_stream: bool = False,
    ):
        self._client_id = str(uuid.uuid1())
        splits = table_name.split("/")
        if len(splits) == 1:
            pass
        elif len(splits) == 2:
            if project_name is not None:
                raise ValueError(
                    f"Cannot specify project_name and table_name with '/' in it: {table_name}"
                )
            project_name = splits[0]
            table_name = splits[1]
        elif len(splits) == 3:
            if project_name is not None or entity_name is not None:
                raise ValueError(
                    f"Cannot specify project_name or entity_name and table_name with 2 '/'s in it: {table_name}"
                )
            entity_name = splits[0]
            project_name = splits[1]
            table_name = splits[2]

        # For now, we force the user to specify the entity and project
        # technically, we could infer the entity from the API key, but
        # that tends to confuse users.
        if entity_name is None or entity_name == "":
            raise ValueError(f"Must specify entity_name")
        elif project_name is None or project_name == "":
            raise ValueError(f"Must specify project_name")
        elif table_name is None or table_name == "":
            raise ValueError(f"Must specify table_name")

        job_type = "wb_stream_table"
        self._lite_run = InMemoryLazyLiteRun(
            entity_name,
            project_name,
            table_name,
            job_type,
            _use_async_file_stream=not _disable_async_file_stream,
        )
        self._table_name = self._lite_run._run_name
        self._project_name = self._lite_run._project_name
        self._entity_name = self._lite_run._entity_name
        self._ensure_remote_initialized()
        atexit.register(self._at_exit)

    def _ensure_remote_initialized(self) -> StreamTableType:
        self._lite_run.ensure_run()
        print_url = False
        if not hasattr(self, "_weave_stream_table"):
            print_url = True
            self._weave_stream_table = StreamTableType(
                table_name=self._table_name,
                project_name=self._project_name,
                entity_name=self._entity_name,
            )
            # TODO: this generates a second run to save the artifact. Would
            # be nice if we could just use the current run.
            self._weave_stream_table_ref = storage._direct_publish(
                self._weave_stream_table,
                name=self._table_name,
                wb_project_name=self._project_name,
                wb_entity_name=self._entity_name,
            )
        if not hasattr(self, "_artifact"):
            uri = runfiles_wandb.WeaveWBRunFilesURI.from_run_identifiers(
                self._entity_name,
                self._project_name,
                self._table_name,
            )
            self._artifact = WandbLiveRunFiles(name=uri.name, uri=uri)
            self._artifact.set_file_pusher(self._lite_run.pusher)
        if print_url:
            url = f"https://weave.wandb.ai/?exp=get%28%0A++++%22wandb-artifact%3A%2F%2F%2F{self._entity_name}%2F{self._project_name}%2F{self._table_name}%3Alatest%2Fobj%22%29%0A++.rows"
            printer = get_printer(_get_python_type() != "python")
            printer.display(f'{printer.emoji("star")} View data at {printer.link(url)}')
        return self._weave_stream_table

    def log(self, row_or_rows: ROW_TYPE) -> None:
        if isinstance(row_or_rows, dict):
            row_or_rows = [row_or_rows]

        for row in row_or_rows:
            self._log_row(row)

    def rows(self) -> graph.Node:
        if self._weave_stream_table_ref is None:
            raise errors.WeaveInternalError("ref is None after ensure")
        return stream_table_ops.rows(
            weave_api.get(str(self._weave_stream_table_ref.uri))
        )

    def _ipython_display_(self) -> graph.Node:
        from .. import show

        return show(self.rows())

    def _log_row(self, row: dict) -> None:
        row_copy = {**row}
        row_copy["_client_id"] = self._client_id
        if "timestamp" not in row_copy:
            row_copy["timestamp"] = datetime.datetime.now()
        payload = row_to_weave(row_copy, self._artifact)
        self._lite_run.log(payload)

    def finish(self) -> None:
        if self._artifact:
            self._artifact.cleanup()
        if self._lite_run:
            self._lite_run.finish()

    def __del__(self) -> None:
        self.finish()

    def _at_exit(self) -> None:
        self.finish()


class StreamTable(_StreamTableSync):
    MAX_UNSAVED_SECONDS = 2

    def __init__(
        self,
        table_name: str,
        *,
        project_name: typing.Optional[str] = None,
        entity_name: typing.Optional[str] = None,
        _disable_async_file_stream: bool = False,
    ):
        super().__init__(
            table_name=table_name,
            project_name=project_name,
            entity_name=entity_name,
            _disable_async_file_stream=_disable_async_file_stream,
        )

        self.queue: queue.Queue = queue.Queue()
        atexit.register(self._at_exit)
        self._lock = threading.Lock()
        self._join_event = threading.Event()
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True
        self._thread.start()

    def log(self, row_or_rows: ROW_TYPE) -> None:
        self.queue.put(row_or_rows)

    def _flush(self) -> None:
        with self._lock:
            for log_payload in self._iterate_queue():
                super().log(log_payload)

    def _iterate_queue(
        self,
    ) -> typing.Generator[ROW_TYPE, None, None]:
        while True:
            try:
                record = self.queue.get_nowait()
            except queue.Empty:
                break
            else:
                yield record
                self.queue.task_done()

    def _thread_body(self) -> None:
        join_requested = False
        while not join_requested:
            join_requested = self._join_event.wait(self.MAX_UNSAVED_SECONDS)
            self._flush()

    # Override methods of _StreamTableSync
    def finish(self) -> None:
        if hasattr(self, "_thread"):
            self._join_event.set()
            self._thread.join()
            with self._lock:
                super().finish()


def maybe_history_type_to_weave_type(tc_type: str) -> typing.Optional[weave_types.Type]:
    if tc_type.startswith(TYPE_ENCODE_PREFIX):
        w_type = json.loads(tc_type[len(TYPE_ENCODE_PREFIX) :])
        return weave_types.TypeRegistry.type_from_dict(w_type)
    else:
        possible_type = weave_types.type_name_to_type(tc_type)
        if possible_type is not None:
            try:
                return possible_type()
            except Exception as e:
                logging.warning(
                    f"StreamTable Type Error: Found type for {tc_type}, but blind construction failed: {e}",
                )
    return None


def is_weave_encoded_history_cell(cell: dict) -> bool:
    return "_val" in cell and (
        "_weave_type" in cell
        or (cell.get("_type") != None and cell["_type"].startswith(TYPE_ENCODE_PREFIX))
    )


def from_weave_encoded_history_cell(cell: dict) -> typing.Any:
    if not is_weave_encoded_history_cell(cell):
        raise ValueError(f"Expected weave encoded history cell, got {cell}")
    if "_weave_type" in cell:
        weave_type = cell["_weave_type"]
    elif cell["_type"].startswith(TYPE_ENCODE_PREFIX):
        weave_type = json.loads(cell["_type"][len(TYPE_ENCODE_PREFIX) :])
    else:
        raise ValueError(f"Expected weave encoded history cell, got {cell}")
    weave_json = {
        "_type": weave_type,
        "_val": cell["_val"],
    }
    return storage.from_python(weave_json)


def row_to_weave(
    row: typing.Dict[str, typing.Any], artifact: WandbLiveRunFiles
) -> typing.Dict[str, typing.Any]:
    return {key: obj_to_weave(value, artifact) for key, value in row.items()}


def obj_to_weave(obj: typing.Any, artifact: WandbLiveRunFiles) -> typing.Any:
    def recurse(obj: typing.Any) -> typing.Any:
        return obj_to_weave(obj, artifact)

    # all primitives
    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj
    else:
        if TRUST_GORILLA_FOR_NESTED_STRUCTURES:
            if isinstance(obj, dict):
                return {key: recurse(value) for key, value in obj.items()}
            elif isinstance(obj, list):
                return [recurse(value) for value in obj]
            elif isinstance(obj, tuple):
                return [recurse(value) for value in obj]
            elif isinstance(obj, set):
                return [recurse(value) for value in obj]
            elif isinstance(obj, frozenset):
                return [recurse(value) for value in obj]
            else:
                return leaf_to_weave(obj, artifact)
        else:
            return leaf_to_weave(obj, artifact)


def w_type_to_type_name(w_type: typing.Union[str, dict]) -> str:
    if isinstance(w_type, str):
        return w_type
    if ENCODE_ENTIRE_TYPE:
        return TYPE_ENCODE_PREFIX + json.dumps(w_type)
    else:
        return w_type["type"]


def leaf_to_weave(leaf: typing.Any, artifact: WandbLiveRunFiles) -> typing.Any:
    def ref_persister_artifact(
        type: weave_types.Type, refs: typing.Iterable[artifact_base.ArtifactRef]
    ) -> artifact_base.Artifact:
        # Save all the reffed objects into the new artifact.
        for mem_ref in refs:
            if mem_ref.path is not None and mem_ref._type is not None:
                # Hack: add a random salt to the end (i really want content addressing here)
                # but this is a quick fix to avoid collisions
                path = mem_ref.path + "-" + str(uuid.uuid4())
                artifact.set(path, mem_ref._type, mem_ref._obj)
        return artifact

    res = storage.to_python(leaf, None, ref_persister_artifact)

    w_type = res["_type"]
    type_name = w_type_to_type_name(w_type)

    if ENCODE_ENTIRE_TYPE:
        return {"_type": type_name, "_val": res["_val"]}
    else:
        return {
            "_type": type_name,
            "_weave_type": w_type,
            "_val": res["_val"],
        }



# Official interface for making http requests. All weave
# interactions with http servers should go through this interface.

import asyncio
import time
import os
import aiohttp
import types
import typing
import yarl
import logging
import requests


from . import engine_trace
from . import filesystem
from . import server_error_handling

logging.getLogger("aiohttp.access").setLevel(logging.WARNING)
logging.getLogger("aiohttp.client").setLevel(logging.WARNING)
logging.getLogger("aiohttp.internal").setLevel(logging.WARNING)
logging.getLogger("aiohttp.server").setLevel(logging.WARNING)
logging.getLogger("aiohttp.web").setLevel(logging.WARNING)
logging.getLogger("aiohttp.websocket").setLevel(logging.WARNING)


tracer = engine_trace.tracer()  # type: ignore

# Turn on to print stats for each request
ENABLE_REQUEST_TRACING = False


def logging_trace_config() -> aiohttp.TraceConfig:
    def make_trace_event_handler(
        event_name: str,
    ) -> typing.Callable[
        [aiohttp.ClientSession, types.SimpleNamespace, typing.Any],
        typing.Coroutine[None, None, None],
    ]:
        async def on_event(
            session: aiohttp.ClientSession,
            trace_config_ctx: types.SimpleNamespace,
            params: typing.Any,
        ) -> None:
            if not hasattr(trace_config_ctx, "events"):
                trace_config_ctx.events = []
            trace_config_ctx.events.append(
                (event_name, asyncio.get_event_loop().time())
            )

        return on_event

    async def on_request_end(
        session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.tracing.TraceRequestEndParams,
    ) -> None:
        trace_config_ctx.events.append(("redirect", asyncio.get_event_loop().time()))
        trace_config_ctx.events.append(("req_end", asyncio.get_event_loop().time()))
        events = trace_config_ctx.events
        prior_e = events[0]
        s = f"{prior_e[0]}"
        for e in events[1:]:
            s += f" [{e[1]-prior_e[1]:.3f}s] {e[0]}"
            prior_e = e
        print("Req done: ", time.time(), s)

    trace_config = aiohttp.TraceConfig()
    trace_config.on_connection_queued_start.append(
        make_trace_event_handler("queued_start")
    )
    trace_config.on_connection_queued_end.append(make_trace_event_handler("queued_end"))
    trace_config.on_connection_create_start.append(
        make_trace_event_handler("conn_create_start")
    )
    trace_config.on_connection_create_end.append(
        make_trace_event_handler("conn_create_end")
    )
    trace_config.on_request_start.append(make_trace_event_handler("request_start"))
    trace_config.on_request_start.append(make_trace_event_handler("request_redirect"))
    trace_config.on_request_end.append(on_request_end)
    return trace_config


class HttpAsync:
    def __init__(self, fs: filesystem.FilesystemAsync) -> None:
        self.fs = fs

        conn = aiohttp.TCPConnector(limit=50)
        trace_configs = []
        if ENABLE_REQUEST_TRACING:
            trace_configs.append(logging_trace_config())
        self.session = aiohttp.ClientSession(
            trace_configs=trace_configs,
            connector=conn,
            cookie_jar=aiohttp.DummyCookieJar(),
        )

    async def __aenter__(self) -> "HttpAsync":
        return self

    async def __aexit__(self, *args: typing.Any) -> None:
        await self.session.close()

    async def download_file(
        self,
        url: str,
        path: str,
        headers: typing.Optional[dict[str, str]] = None,
        cookies: typing.Optional[dict[str, str]] = None,
        auth: typing.Optional[aiohttp.BasicAuth] = None,
    ) -> None:
        await self.fs.makedirs(os.path.dirname(path), exist_ok=True)
        with tracer.trace("download_file_task"):
            # TODO: Error handling when no file or manifest

            # yarl.URL encoded=True is very important! Otherwise aiohttp
            # will encode the url again and we'll get a 404 for things like
            # signed URLs
            async with self.session.get(
                yarl.URL(url, encoded=True), headers=headers, cookies=cookies, auth=auth
            ) as r:
                if r.status == 200:
                    async with self.fs.open_write(path, mode="wb") as f:
                        async for data in r.content.iter_chunked(16 * 1024):
                            await f.write(data)
                else:
                    raise server_error_handling.WeaveInternalHttpException.from_code(
                        r.status, "Download failed"
                    )


class Http:
    def __init__(self, fs: filesystem.Filesystem) -> None:
        self.fs = fs
        self.session = requests.Session()

    def __enter__(self) -> "Http":
        return self

    def __exit__(self, *args: typing.Any) -> None:
        self.session.close()

    def download_file(
        self,
        url: str,
        path: str,
        headers: typing.Optional[dict[str, str]] = None,
        cookies: typing.Optional[dict[str, str]] = None,
    ) -> None:
        self.fs.makedirs(os.path.dirname(path), exist_ok=True)
        with tracer.trace("download_file_task"):
            # TODO: Error handling when no file or manifest

            # yarl.URL encoded=True is very important! Otherwise aiohttp
            # will encode the url again and we'll get a 404 for things like
            # signed URLs
            with self.session.get(
                str(yarl.URL(url, encoded=True)), headers=headers, cookies=cookies
            ) as r:
                if r.status_code == 200:  # type: ignore
                    with self.fs.open_write(path, mode="wb") as f:
                        f.write(r.content)  # type: ignore
                else:
                    raise server_error_handling.WeaveInternalHttpException.from_code(
                        r.status_code, "Download failed"  # type: ignore
                    )



import weave

from . import slackapi_readexport


@weave.type()
class Message:
    ts: str
    channel: str
    text: str
    user_id: str


@weave.type()
class Channel:
    # TODO: this should be SlackApi, but serialization doesn't work when we use a Protocol here :(
    slack_api: slackapi_readexport.SlackReadExportApi
    channel_name: str  # can't use 'name' because of VarNode attr conflict :(

    @weave.op()
    def size(self) -> int:
        return self.slack_api.channel_export_size(self.channel_name)

    @weave.op()
    def messages(self) -> list[Message]:
        channel_messages = []
        for message in self.slack_api.channel_messages(self.channel_name):
            if message["type"] != "message" or message.get("subtype"):
                # TODO: Don't drop these! Should be exposed as part of the API and filtered out.

                # actual messages from users don't have subtype set
                continue
            channel_messages.append(
                Message(
                    message["ts"], self.channel_name, message["text"], message["user"]
                )
            )
        return channel_messages


@weave.type()
class Slack:
    # TODO: this should be SlackApi, but serialization doesn't work when we use a Protocol here :(
    slack_api: slackapi_readexport.SlackReadExportApi

    @weave.op()
    def channels(self) -> list[Channel]:
        # TODO: sorted?
        return [
            self.channel.resolve_fn(self, name)
            for name in self.slack_api.channel_names()
        ]

    @weave.op()
    def channel(self, name: str) -> Channel:
        # TODO: pass in data_dir
        return Channel(self.slack_api, name)


# Marked as impure, because we need to read the directory d on every execution.
# TODO: This should not need to be manually marked. We know that VersionedDir reads
#     a local dir. One way to fix this would be to use Weave ops for constructors
#     instead of Python class constructors (ie we'd have a versioned_dir op that is
#     impure, and everything downstream of it would be impure until we hit an existing
#     artifact version).
@weave.op(render_info={"type": "function"}, pure=False)
def open_slack_export(d: str) -> Slack:
    # Note from Shawn: I changed this because I got rid of "VersionedDir". Now passing
    # Dir which is more like DirInfo. Don't know if this still works.
    return Slack(slackapi_readexport.SlackReadExportApi(weave.Dir(d, 0, {}, {})))


def all_messages(channels: list[Channel]) -> list[Message]:
    ms = []
    for c in channels:
        ms += c.messages
    return ms



import os
import typing

from ..api import op
from .. import weave_types as types
from .. import file_local
from .. import environment


@op(name="LocalFile-directUrl")
def direct_url(file: file_local.LocalFile) -> str:
    local_path = os.path.abspath(file.get_local_path())
    return f"{environment.weave_server_url()}/__weave/file{local_path}"


@op(name="LocalFile-directUrlAsOf")
def direct_url_as_of(file: file_local.LocalFile, asOf: int) -> str:
    local_path = os.path.abspath(file.get_local_path())
    return f"{environment.weave_server_url()}/__weave/file{local_path}"


@op(name="localpathReturnType", hidden=True)
def local_path_refine_type(path: str) -> types.Type:
    return file_local.get_path_type(path)


@op(
    name="localpath",
    pure=False,
    render_info={"type": "function"},
    refine_output_type=local_path_refine_type,
)
def local_path(
    path: str,
) -> typing.Optional[typing.Union[file_local.LocalDir, file_local.LocalFile]]:
    return file_local.get_path_info(path)



from __future__ import annotations
from typing_extensions import (
    NotRequired,
    TypedDict,
)
import typing
import logging
import os

import tiktoken
import numpy as np
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
import openai

import weave
from ... import ops_arrow

logger = logging.getLogger(__name__)
openai.api_key = os.environ.get("OPENAI_API_KEY")

# OpenAI API functions
retry_openai_decorator = retry(
    reraise=True,
    stop=stop_after_attempt(4),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    retry=(
        retry_if_exception_type(openai.error.Timeout)
        | retry_if_exception_type(openai.error.APIError)
        | retry_if_exception_type(openai.error.APIConnectionError)
        | retry_if_exception_type(openai.error.RateLimitError)
        | retry_if_exception_type(openai.error.ServiceUnavailableError)
    ),
    before_sleep=before_sleep_log(logger, logging.WARNING),
)


@retry_openai_decorator
def _openai_embed(model, input):
    return openai.Embedding.create(input=input, model=model)


# @retry_openai_decorator
# def openai_chatcompletion(model, messages):
#     return openai.ChatCompletion.create(
#         model="gpt-3.5-turbo",  # The deployment name you chose when you deployed the ChatGPT or GPT-4 model.
#         messages=messages,
#     )


# Helper to efficiently embed a set of documents using the OpenAI embedding API
# This is from langchain

embedding_ctx_length = 8191
OPENAI_EMBEDDING_MODEL = "text-embedding-ada-002"
chunk_size = 1000


# Based on the langchain implementation
def embed_texts(texts: list[str], embedding_model: str) -> list[list[float]]:
    embeddings: list[list[float]] = [[] for _ in range(len(texts))]
    tokens = []
    indices = []
    encoding = tiktoken.model.encoding_for_model(embedding_model)
    for i, text in enumerate(texts):
        if embedding_model.endswith("001"):
            # See: https://github.com/openai/openai-python/issues/418#issuecomment-1525939500
            # replace newlines, which can negatively affect performance.
            text = text.replace("\n", " ")
        token = encoding.encode(
            text,
            disallowed_special="all",
        )
        for j in range(0, len(token), embedding_ctx_length):
            tokens += [token[j : j + embedding_ctx_length]]
            indices += [i]

    batched_embeddings = []
    _chunk_size = chunk_size
    for i in range(0, len(tokens), _chunk_size):
        response = _openai_embed(
            embedding_model,
            input=tokens[i : i + _chunk_size],
        )
        batched_embeddings += [r["embedding"] for r in response["data"]]

    results: list[list[list[float]]] = [[] for _ in range(len(texts))]
    num_tokens_in_batch: list[list[int]] = [[] for _ in range(len(texts))]
    for i in range(len(indices)):
        results[indices[i]].append(batched_embeddings[i])
        num_tokens_in_batch[indices[i]].append(len(tokens[i]))

    for i in range(len(texts)):
        _result = results[i]
        if len(_result) == 0:
            average = _openai_embed(embedding_model, input="",)["data"][
                0
            ]["embedding"]
        else:
            average = np.average(_result, axis=0, weights=num_tokens_in_batch[i])
        embeddings[i] = (average / np.linalg.norm(average)).tolist()

    return embeddings


class OpenAIEmbedOptions(TypedDict):
    model: NotRequired[str]


@weave.op()
def openai_embed(
    texts: list[typing.Optional[str]], options: OpenAIEmbedOptions
) -> ops_arrow.ArrowWeaveList[list[float]]:
    model = typing.cast(str, options.get("model", OPENAI_EMBEDDING_MODEL))
    non_none_texts: list[str] = []
    result_positions: list[typing.Optional[int]] = []
    for i, text in enumerate(texts):
        if text is None or text == None:
            result_positions.append(None)
        else:
            result_positions.append(len(non_none_texts))
            non_none_texts.append(text)
    res = embed_texts(non_none_texts, model)
    final_result: list[typing.Optional[list[float]]] = []
    for pos in result_positions:
        if pos is None:
            final_result.append(None)
        else:
            final_result.append(res[pos])
    return ops_arrow.to_arrow(final_result)



from ..api import op
from .. import weave_types as types

from .arrow import ArrowWeaveListType

from . import convert


@op(
    input_type={
        "arr": types.List(),
    },
    output_type=lambda input_types: ArrowWeaveListType(input_types["arr"].object_type),
)
def list_to_arrow(arr):
    return convert.to_arrow(arr)


@op(
    input_type={
        "self": types.UnionType(ArrowWeaveListType(), types.List()),
    },
    output_type=lambda input_types: types.List(input_types["self"].object_type),
)
def to_py(self):
    if isinstance(self, types.List):
        return self
    return self.to_pylist_tagged()



import datetime

import dataclasses
import io
import plotly
import weave
import typing
import plotly.express as px
import pandas as pd
from plotly import graph_objs as go

from ... import weave_internal
from ... import infer_types


class PlotlyType(weave.types.Type):
    instance_classes = plotly.graph_objs.Figure

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.plotly") as f:
            plotly.io.write_json(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.plotly") as f:
            return plotly.io.from_json(f.read())


@weave.weave_class(weave_type=PlotlyType)
class PlotlyOps:
    @weave.op()
    def contents(self) -> str:
        f = io.StringIO()
        plotly.io.write_json(self, f)
        f.seek(0)
        return f.read()


class BoxRange(typing.TypedDict):
    xMin: float
    xMax: float
    yMin: float
    yMax: float


@weave.type()
class PanelPlotlyConfig:
    selected: weave.Node[BoxRange] = dataclasses.field(
        default_factory=lambda: weave_internal.make_const_node(
            infer_types.python_type_to_type(BoxRange),
            {"xMin": 0, "xMax": 0, "yMin": 0, "yMax": 0},
        )
    )


@weave.type()
class PanelPlotly(weave.Panel):
    id = "PanelPlotly"
    config: PanelPlotlyConfig = dataclasses.field(default_factory=PanelPlotlyConfig)


class BarData:
    value: float
    label: str
    count: float


@weave.op()
def plotly_barplot(bar_data: list[BarData]) -> plotly.graph_objs.Figure:
    bar_data = list(bar_data)  # do this to get out of arrow :(
    if not bar_data:
        fig = plotly.graph_objs.Figure()
        fig.update_layout(template="plotly_white")
        return fig
    fig = px.bar(
        bar_data,
        x="value",
        y="count",
        color="label",
        barmode="group",
        template="plotly_white",
    )
    fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
    fig.update_layout(dragmode="select")
    return fig


class ScatterData(typing.TypedDict):
    x: typing.Optional[float]
    y: typing.Optional[float]
    label: typing.Optional[str]


class TimeBin(typing.TypedDict):
    start: datetime.datetime
    center: datetime.datetime
    stop: datetime.datetime


class TimeSeriesData(typing.TypedDict):
    x: TimeBin
    y: float
    label: str


@weave.op(
    input_type={
        "data": weave.types.List(
            weave.types.TypedDict(
                {
                    "x": weave.types.TypedDict(
                        {
                            key: weave.types.optional(weave.types.Timestamp())
                            for key in ["start", "stop"]
                        }
                    ),
                    "y": weave.types.optional(weave.types.Float()),
                    "label": weave.types.optional(weave.types.String()),
                }
            )
        ),
        "mark": weave.types.String(),
        "labels": weave.types.TypedDict(
            {
                "x": weave.types.String(),
                "y": weave.types.String(),
                "label": weave.types.String(),
            }
        ),
        "label_overrides": weave.types.Dict(weave.types.String(), weave.types.String()),
    }
)
def plotly_time_series(data, mark, labels, label_overrides) -> plotly.graph_objs.Figure:
    data = list(data)  # do this to get out of arrow :(
    labels = {**labels, **label_overrides}

    def bin_center(bin):
        return bin["x"]["start"] + (bin["x"]["stop"] - bin["x"]["start"]) / 2

    if mark == "point":
        data = [{**d, "x": bin_center(d)} for d in data]  # type: ignore
        fig = px.scatter(
            data, x="x", y="y", color="label", template="plotly_white", labels=labels
        )
        fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
        fig.update_layout(dragmode="select")
        return fig
    elif mark == "bar":
        pruned = [{**d, "x": d["x"]["start"]} for d in data]  # type: ignore

        # plotly expects timedeltas in ms.
        # source https://community.plotly.com/t/bar-width-using-plotly-express/47580/5
        bin_size = (data[0]["x"]["stop"] - data[0]["x"]["start"]).total_seconds() * 1000
        fig = px.bar(
            pruned,
            x="x",
            y="y",
            color="label",
            barmode="stack",
            template="plotly_white",
            labels=labels,
        )
        fig.update_layout(showlegend=False)
        fig.update_traces(width=bin_size)
        fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
        fig.update_layout(dragmode="select")
        return fig
    elif mark == "line":
        data = [{**d, "x": bin_center(d)} for d in data]  # type: ignore
        fig = px.line(
            data, x="x", y="y", color="label", template="plotly_white", labels=labels
        )
        fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
        fig.update_layout(dragmode="select")
        return fig
    else:
        raise ValueError(f"Unknown mark {mark}")


@weave.op()
def plotly_scatter(data: list[ScatterData]) -> plotly.graph_objs.Figure:
    from ... import storage

    data = storage.to_weavejs(data)
    color = None
    if data and "label" in data[0]:
        color = "label"
    fig = px.scatter(data, x="x", y="y", color=color, template="plotly_white")
    fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
    fig.update_layout(dragmode="select")
    return fig


class GeoData(typing.TypedDict):
    lat: float
    long: float
    color: typing.Union[float, str]


@weave.op()
def plotly_geo(data: list[GeoData]) -> plotly.graph_objs.Figure:
    de = pd.DataFrame(data)

    fig = go.Figure(
        data=go.Scattergeo(
            lon=de["long"],
            lat=de["lat"],
            mode="markers",
            # text=de["Map label"],  # str(de["Magnitude"]) + " " + de["Date"],
            marker_color=de["color"]  # , colors[df["Type"][0]]
            # showlegend=True,
            # marker=dict(
            #     color=de["Magnitude"], size=15, opacity=0.9, colorscale="Sunset"
            # )
            # opacity
            # marker_color=colors[volcano_types[df["Type"][0]]]
        )
    )
    fig.update_geos(
        projection_type="orthographic",
        landcolor="white",
        oceancolor="MidnightBlue",
        showocean=True,
        lakecolor="LightBlue",
    )
    return fig



import copy
import math
import functools

from ..api import op, weave_class
from .. import decorator_type
from .. import weave_types as types
from . import list_
from . import graph

from ..language_features.tagging import tagged_value_type

try:
    import sqlalchemy
    from sqlalchemy import orm
except ImportError:
    pass


@decorator_type.type("sqlconnection")
class SqlConnection:
    path: str

    @functools.cached_property
    def engine(self):
        return sqlalchemy.create_engine(self.path)

    @functools.cached_property
    def meta(self):
        meta = sqlalchemy.MetaData()
        meta.reflect(self.engine)
        return meta

    def table(self, name):
        return SqlTable(self, name)


def filter_fn_to_sql_filter(table, filter_fn_node):
    if isinstance(filter_fn_node, graph.ConstNode):
        return filter_fn_node.val
    elif isinstance(filter_fn_node, graph.OutputNode):
        op_name = graph.op_full_name(filter_fn_node.from_op)
        if op_name == "number-greater":
            return filter_fn_to_sql_filter(
                table, filter_fn_node.from_op.inputs["lhs"]
            ) > filter_fn_to_sql_filter(table, filter_fn_node.from_op.inputs["rhs"])
        elif op_name == "pick":
            return filter_fn_to_sql_filter(
                table, filter_fn_node.from_op.inputs["obj"]
            ).columns[
                filter_fn_to_sql_filter(table, filter_fn_node.from_op.inputs["key"])
            ]
        elif op_name == "typedDict-pick":
            return filter_fn_to_sql_filter(
                table, filter_fn_node.from_op.inputs["self"]
            ).columns[
                filter_fn_to_sql_filter(table, filter_fn_node.from_op.inputs["key"])
            ]
        raise Exception("unhandled op name", op_name)
    elif isinstance(filter_fn_node, graph.VarNode):
        if filter_fn_node.name == "row":
            return table
        raise Exception("unhandled var name")


def index_output_type(input_types):
    # THIS IS NO GOOD
    # TODO: need to fix Const type so we don't need this.
    self_type = input_types["self"]
    if isinstance(self_type, types.Const):
        return self_type.val_type.object_type
    else:
        return self_type.object_type


def mapped_pick_output_type(input_types):
    if not isinstance(input_types["key"], types.Const):
        return types.List(types.UnknownType())
    key = input_types["key"].val
    prop_type = input_types["self"].object_type.property_types.get(key)
    if prop_type is None:
        return types.Invalid()
    return types.List(prop_type)


class SqlTableType(types.ObjectType):
    _base_type = types.List
    name = "sqltable"

    object_type: types.Type

    def property_types(self):
        return {
            "conn": SqlConnection.WeaveType(),
            "table_name": types.String(),
        }

    def __init__(self, object_type=types.TypedDict({})):
        self.object_type = object_type

    def _to_dict(self):
        return {"objectType": self.object_type.to_dict()}

    @classmethod
    def from_dict(cls, d):
        return cls(types.TypeRegistry.type_from_dict(d["objectType"]))


@weave_class(weave_type=SqlTableType)
class SqlTable:
    PAGE_SIZE = 100

    def __init__(self, conn, table_name):
        self.conn = conn
        self.table_name = table_name

        # We always fetch results in pages of PAGE_SIZE, and store the results
        # here in _row_cache.
        self._row_cache = {}
        self._filter_fn = None

    @functools.cached_property
    def table(self):
        return self.conn.meta.tables[self.table_name]

    def _to_list_table(self):
        self_list = []
        for i in range(self._count()):
            self_list.append(self._index(i))
        return self_list

    def copy(self):
        new_obj = self.__class__(self.conn, self.table_name)
        new_obj._filter_fn = copy.deepcopy(self._filter_fn)
        return new_obj

    def _cached_row(self, index):
        page = math.floor(index / self.PAGE_SIZE)
        page_offset = index % self.PAGE_SIZE
        if page in self._row_cache:
            try:
                return self._row_cache[page][page_offset]
            except IndexError:
                return None
        return None

    @property
    def query(self):
        DBSession = orm.sessionmaker(bind=self.conn.engine)
        session = DBSession()
        query = session.query(self.table)
        if self._filter_fn is not None:
            query = query.filter(filter_fn_to_sql_filter(self.table, self._filter_fn))
        return query

    def _count(self):
        return self.query.count()

    @op()
    def count(self) -> int:
        return self._count()

    def _index(self, index: int):
        row = self._cached_row(index)
        if row is not None:
            return row

        page = math.floor(index / self.PAGE_SIZE)
        page_offset = index % self.PAGE_SIZE

        query = self.query
        results = query.offset(page * self.PAGE_SIZE).limit(SqlTable.PAGE_SIZE)
        # print('RESULTS QUERY', results)
        rows = []
        for row in results.all():
            row = {k: getattr(row, k) for k in row._fields}
            rows.append(row)
        self._row_cache[page] = rows
        try:
            return rows[page_offset]
        except IndexError:
            return None

    @op(name="sqltable-index", output_type=index_output_type)
    def __getitem__(self, index: int):
        return self._index(index)

    @op(output_type=mapped_pick_output_type)
    def pick(self, key: str):
        return list_.general_picker(self._to_list_table(), key)

    @op(
        input_type={
            "map_fn": lambda input_types: types.Function(
                {"row": input_types["self"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: types.List(input_types["self"].object_type),
    )
    def map(self, map_fn):
        return list_.List.map.resolve_fn(self._to_list_table(), map_fn)

    @op(
        input_type={
            "filterFn": lambda input_types: types.Function(
                {"row": input_types["self"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: input_types["self"],
    )
    def filter(self, filterFn):
        new_obj = self.copy()
        new_obj._filter_fn = filterFn
        return new_obj

    @op(
        input_type={
            "group_by_fn": lambda input_types: types.Function(
                {"row": input_types["self"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: types.List(
            tagged_value_type.TaggedValueType(
                types.TypedDict(
                    {
                        "groupKey": input_types["group_by_fn"].output_type,
                    }
                ),
                types.List(input_types["self"].object_type),
            )
        ),
    )
    def groupby(self, group_by_fn):
        return list_.List.groupby.resolve_fn(self._to_list_table(), group_by_fn)


SqlTableType.instance_class = SqlTable
SqlTableType.instance_classes = SqlTable


@op(
    name="local-sqlconnection",
    input_type={"path": types.String()},
    output_type=SqlConnection.WeaveType(),  # type: ignore
)
def local_sqlconnection(path):
    return SqlConnection(path)


@op(
    name="sqlconnection-tables",
    input_type={"conn": SqlConnection.WeaveType()},  # type: ignore
    output_type=types.TypedDict({}),
)
def sqlconnection_tables(conn):
    tables = {}
    for table_name in conn.meta.tables.keys():
        tables[table_name] = conn.table(table_name)
    return tables


# @op(
#     name='sqlconnection-schemanames',
#     input_type={
#         'conn': SqlConnectionType()},
#     output_type=types.List(types.String()))
# def sqlconnection_schemanames(conn):
#     return conn.schema_names()

# @op(
#     name='sqlconnection-tablenames',
#     input_type={
#         'conn': SqlConnectionType()},
#     output_type=types.List(types.String()))
# def sqlconnection_tablenames(conn):
#     return conn.table_names()


def _table_to_type(table):
    table_column_types = {}
    for column_name, column in table.columns.items():
        type_class = column.type.__class__.__name__
        if type_class == "VARCHAR" or type_class == "TEXT":
            table_column_type = types.String()
        elif type_class == "INTEGER":
            table_column_type = types.Number()
        table_column_types[column_name] = table_column_type
    return table_column_types


@op(
    name="sqlconnection-tablesType",
    input_type={"conn": SqlConnection.WeaveType()},  # type: ignore
    output_type=types.TypeType(),
)
def sqlconnection_tables_type(conn):
    properties = {}
    for table_name, table in conn.meta.tables.items():
        properties[table_name] = types.List(types.TypedDict(_table_to_type(table)))
    return types.TypedDict(properties).to_json()


@op(
    name="sqlconnection-refine_table",
    hidden=True,
    input_type={"conn": SqlConnection.WeaveType(), "name": types.String()},  # type: ignore
    output_type=types.TypeType(),
)
def refine_sqlconnection_table(conn, name):
    for table_name, table in conn.meta.tables.items():
        if table_name == name:
            return SqlTableType(types.TypedDict(_table_to_type(table)))
    return types.NoneType()


@op(
    name="sqlconnection-table",
    input_type={"conn": SqlConnection.WeaveType(), "name": types.String()},  # type: ignore
    output_type=SqlTableType(types.TypedDict({})),
    refine_output_type=refine_sqlconnection_table,
)
def sqlconnection_table(conn, name):
    return conn.table(name)



import dataclasses
import datetime
import inspect
import math

import pandas as pd

from . import mappers
from . import storage
from . import ref_base
from . import mappers_weave
from . import weave_types as types
from . import errors
from . import box
from . import mappers_python
from . import val_const
from . import artifact_fs
from . import timestamp as weave_timestamp
from .language_features.tagging import tagged_value_type


class TypedDictToPyDict(mappers_weave.TypedDictMapper):
    def apply(self, obj):
        result = {}
        for k, prop_serializer in self._property_serializers.items():
            result[k] = prop_serializer.apply(obj.get(k, None))
        return result


class DictToPyDict(mappers_weave.DictMapper):
    def apply(self, obj):
        result = {}
        for k, v in obj.items():
            k = self.key_serializer.apply(k)
            v = self.value_serializer.apply(v)
            result[k] = v
        return result


class ObjectToPyDict(mappers_weave.ObjectMapper):
    def apply(self, obj):
        # Store the type name in the saved object. W&B for example stores
        # {"_type": "table-ref", "artifact_path": "..."} objects in run config and summary
        # fields.
        result = {"_type": self.type.name}
        for prop_name, prop_serializer in self._property_serializers.items():
            if prop_serializer is not None:
                v = prop_serializer.apply(getattr(obj, prop_name))
                result[prop_name] = v
        return result


class ObjectDictToObject(mappers_weave.ObjectMapper):
    def apply(self, obj):
        # Only add keys that are accepted by the constructor.
        # This is used for Panels where we have an Class-level id attribute
        # that we want to include in the serialized representation.
        result = {}
        result_type = self._obj_type

        # TODO: I think these are hacks in my branch. What do they do?
        instance_class = result_type._instance_classes()[0]
        constructor_sig = inspect.signature(instance_class)
        for k, serializer in self._property_serializers.items():
            if k in constructor_sig.parameters:
                # None haxxx
                # TODO: remove
                obj_val = obj.get(k)
                if obj_val is not None:
                    v = serializer.apply(obj_val)
                    result[k] = v

        for prop_name, prop_type in result_type.type_vars.items():
            if isinstance(prop_type, types.Const):
                result[prop_name] = prop_type.val

        if "artifact" in constructor_sig.parameters and "artifact" not in result:
            result["artifact"] = self._artifact
        try:
            return instance_class(**result)
        except:
            raise errors.WeaveSerializeError(
                "Failed to construct %s with %s" % (instance_class, result)
            )


class ListToPyList(mappers_weave.ListMapper):
    def apply(self, obj):
        return [self._object_type.apply(item) for item in obj]


class UnionToPyUnion(mappers_weave.UnionMapper):
    def apply(self, obj):
        obj_type = types.TypeRegistry.type_of(obj)
        for i, (member_type, member_mapper) in enumerate(
            zip(self.type.members, self._member_mappers)
        ):
            # TODO: Should types.TypeRegistry.type_of always return a const type??
            if isinstance(member_type, types.Const) and not isinstance(
                obj_type, types.Const
            ):
                obj_type = types.Const(obj_type, obj)

            # TODO: assignment isn't right here (a dict with 'a', 'b' int keys is
            # assignable to a dict with an 'a' int key). We want type equality.
            # But that breaks some stuff
            #
            # Later (3/8/23): This is even more ridiculous now. First we check
            # assignablility, then mergability. The ecosystem notebook now fails
            # without the merge check.
            if member_type.assign_type(obj_type) or not isinstance(
                types.merge_types(obj_type, member_type), types.UnionType
            ):
                result = member_mapper.apply(obj)
                if isinstance(result, dict):
                    result["_union_id"] = i
                else:
                    result = {"_union_id": i, "_val": result}
                return result
        raise Exception("invalid %s" % obj)


class PyUnionToUnion(mappers_weave.UnionMapper):
    def apply(self, obj):
        # Another hack for dealing with lack of union support in weavejs.
        if self.is_single_object_nullable and obj is None:
            return None

        try:
            has_union_id = "_union_id" in obj
        except TypeError:
            has_union_id = False

        # hack for deserializing instances of optional types from JS
        # todo: update JS to handle union_id so this is not needed
        if self.is_single_object_nullable and not has_union_id:
            if obj is None:
                return None
            else:
                non_null_mapper = next(
                    filter(
                        lambda m: not types.NoneType().assign_type(m.type),
                        self._member_mappers,
                    )
                )
                return non_null_mapper.apply(obj)
        member_index = obj["_union_id"]
        if "_val" in obj:
            obj = obj["_val"]
        else:
            obj.pop("_union_id")
        return self._member_mappers[member_index].apply(obj)


class IntToPyInt(mappers.Mapper):
    def apply(self, obj):
        return obj


class BoolToPyBool(mappers.Mapper):
    def apply(self, obj):
        if isinstance(obj, box.BoxedBool):
            return obj.val
        return obj


class FloatToPyFloat(mappers.Mapper):
    def apply(self, obj):
        if math.isnan(obj):
            return "nan"
        return obj


class PyFloatToFloat(mappers.Mapper):
    def apply(self, obj):
        if isinstance(obj, str):
            if obj == "nan":
                return float("nan")
        return obj


class StringToPyString(mappers.Mapper):
    def apply(self, obj):
        return obj


class TimestampToPyTimestamp(mappers.Mapper):
    def apply(self, obj: datetime.datetime):
        # TODO: I have no idea why, but Eval.ipynb ends up with a pandas
        # Timestamp here. This is clearly not the correct fix, but we need to
        # get CI working again.
        if isinstance(obj, pd.Timestamp):
            obj = obj.to_pydatetime()
        return weave_timestamp.python_datetime_to_ms(obj)


class PyTimestampToTimestamp(mappers.Mapper):
    def apply(self, obj):
        # This is here to support the legacy "date" type, the frontend passes
        # RFC 3339 formatted strings
        if isinstance(obj, dict):
            if obj.get("type") != "date":
                raise errors.WeaveInternalError(
                    f'expected object with type date but got "{obj}"'
                )
            return datetime.datetime.strptime(obj["val"], "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            return weave_timestamp.ms_to_python_datetime(obj)


class NoneToPyNone(mappers.Mapper):
    def apply(self, obj):
        return None


class UnknownToPyUnknown(mappers.Mapper):
    def apply(self, obj):
        # This should never be called. Unknown for the object type
        # of empty lists
        raise Exception("invalid %s" % obj)


class RefToPyRef(mappers_weave.RefMapper):
    def __init__(
        self, type_: types.Type, mapper, artifact, path=[], use_stable_refs=True
    ):
        super().__init__(type_, mapper, artifact, path)
        self._use_stable_refs = use_stable_refs

    def apply(self, obj: ref_base.Ref):
        try:
            if self._use_stable_refs:
                return obj.uri
            else:
                return obj.initial_uri
        except NotImplementedError:
            raise errors.WeaveSerializeError('Cannot serialize ref "%s"' % obj)


class PyRefToRef(mappers_weave.RefMapper):
    def apply(self, obj):
        return ref_base.Ref.from_str(obj)


class TypeToPyType(mappers.Mapper):
    def apply(self, obj):
        return obj.to_dict()


class PyTypeToType(mappers.Mapper):
    def apply(self, obj):
        return types.TypeRegistry.type_from_dict(obj)


class ConstToPyConst(mappers_weave.ConstMapper):
    def apply(self, obj):
        val = obj
        if isinstance(obj, val_const.Const):
            return val.val
        return self._val_mapper.apply(obj)


class DefaultToPy(mappers.Mapper):
    def __init__(
        self, type_: types.Type, mapper, artifact, path=[], use_stable_refs=True
    ):
        self.type = type_
        self._artifact = artifact
        self._path = path
        self._row_id = 0
        self._use_stable_refs = use_stable_refs

    def apply(self, obj):
        try:
            return self.type.instance_to_dict(obj)
        except NotImplementedError:
            pass
        # If the ref exists elsewhere, just return its uri.
        # TODO: This doesn't deal with MemArtifactRef!
        existing_ref = storage._get_ref(obj)
        if isinstance(existing_ref, artifact_fs.FilesystemArtifactRef):
            if existing_ref.is_saved:
                if self._use_stable_refs:
                    uri = existing_ref.uri
                else:
                    uri = existing_ref.initial_uri
                return str(uri)
        ref = None
        if isinstance(obj, ref_base.Ref):
            ref = obj
        elif isinstance(obj, str):
            try:
                ref = ref_base.Ref.from_str(obj)
            except (errors.WeaveInternalError, NotImplementedError):
                pass
        if ref is None:
            # This defines the artifact layout!
            name = "/".join(self._path + [str(self._row_id)])
            self._row_id += 1

            ref = self._artifact.set(name, self.type, obj)
        if ref.artifact == self._artifact:
            return ref.local_ref_str()
        else:
            return str(ref)


class DefaultFromPy(mappers.Mapper):
    def __init__(self, type_: types.Type, mapper, artifact, path=[]):
        self.type = type_
        self._artifact = artifact
        self._path = path

    def apply(self, obj):
        if isinstance(obj, dict):
            return self.type.instance_from_dict(obj)
        # else its a ref string
        # TODO: this does not use self.artifact, can we just drop it?
        # Do we need the type so we can load here? No...
        if ":" in obj:
            ref = ref_base.Ref.from_str(obj)
            # Note: we are forcing type here, because we already know it
            # We don't save the types for every file in a remote artifact!
            # But you can still reference them, because you have to get that
            # file through an op, and therefore we know the type?
            ref._type = self.type
            return ref.get()
        return self._artifact.get(obj, self.type)


py_type = type


@dataclasses.dataclass
class RegisteredMapper:
    type_class: type[types.Type]
    to_mapper: type[mappers.Mapper]
    from_mapper: type[mappers.Mapper]


_additional_mappers: list[RegisteredMapper] = []


def add_mapper(
    type_class: type[types.Type],
    to_mapper: type[mappers.Mapper],
    from_mapper: type[mappers.Mapper],
):
    _additional_mappers.append(RegisteredMapper(type_class, to_mapper, from_mapper))


def map_to_python_(type, mapper, artifact, path=[], mapper_options=None):
    mapper_options = mapper_options or {}
    use_stable_refs = mapper_options.get("use_stable_refs", True)
    if isinstance(type, types.TypeType):
        # If we're actually serializing a type itself
        return TypeToPyType(type, mapper, artifact, path)
    elif isinstance(type, types.TypedDict):
        return TypedDictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.Dict):
        return DictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.List):
        return ListToPyList(type, mapper, artifact, path)
    elif isinstance(type, types.UnionType):
        return UnionToPyUnion(type, mapper, artifact, path)
    elif isinstance(type, types.ObjectType):
        return ObjectToPyDict(type, mapper, artifact, path)
    elif isinstance(type, tagged_value_type.TaggedValueType):
        return tagged_value_type.TaggedValueToPy(type, mapper, artifact, path)
    elif isinstance(type, types.Boolean):
        return BoolToPyBool(type, mapper, artifact, path)
    elif isinstance(type, types.Int):
        return IntToPyInt(type, mapper, artifact, path)
    elif isinstance(type, types.Float):
        return FloatToPyFloat(type, mapper, artifact, path)
    elif isinstance(type, types.Number):
        return FloatToPyFloat(type, mapper, artifact, path)
    elif isinstance(type, types.String):
        return StringToPyString(type, mapper, artifact, path)
    elif isinstance(type, types.Timestamp):
        return TimestampToPyTimestamp(type, mapper, artifact, path)
    elif isinstance(type, types.Const):
        return ConstToPyConst(type, mapper, artifact, path)
    elif isinstance(type, types.NoneType):
        return NoneToPyNone(type, mapper, artifact, path)
    elif isinstance(type, types.UnknownType):
        return UnknownToPyUnknown(type, mapper, artifact, path)
    elif isinstance(type, types.RefType):
        return RefToPyRef(type, mapper, artifact, path, use_stable_refs=use_stable_refs)
    else:
        for m in _additional_mappers:
            if isinstance(type, m.type_class):
                return m.to_mapper(type, mapper, artifact, path)
        return DefaultToPy(
            type, mapper, artifact, path, use_stable_refs=use_stable_refs
        )


def map_from_python_(type: types.Type, mapper, artifact, path=[], mapper_options=None):
    if isinstance(type, types.TypeType):
        # If we're actually serializing a type itself
        return PyTypeToType(type, mapper, artifact, path)
    elif isinstance(type, types.ObjectType):
        return ObjectDictToObject(type, mapper, artifact, path)
    elif isinstance(type, types.TypedDict):
        return TypedDictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.Dict):
        return DictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.List):
        return ListToPyList(type, mapper, artifact, path)
    elif isinstance(type, types.UnionType):
        return PyUnionToUnion(type, mapper, artifact, path)
    elif isinstance(type, tagged_value_type.TaggedValueType):
        return tagged_value_type.TaggedValueFromPy(type, mapper, artifact, path)
    elif isinstance(type, types.Boolean):
        return BoolToPyBool(type, mapper, artifact, path)
    elif isinstance(type, types.Int):
        return IntToPyInt(type, mapper, artifact, path)
    elif isinstance(type, types.Float):
        return PyFloatToFloat(type, mapper, artifact, path)
    elif isinstance(type, types.Number):
        return PyFloatToFloat(type, mapper, artifact, path)
    elif isinstance(type, types.String):
        return StringToPyString(type, mapper, artifact, path)
    elif isinstance(type, types.Timestamp):
        return PyTimestampToTimestamp(type, mapper, artifact, path)
    elif isinstance(type, types.LegacyDate):
        return PyTimestampToTimestamp(type, mapper, artifact, path)
    elif isinstance(type, types.Const):
        return ConstToPyConst(type, mapper, artifact, path)
    elif isinstance(type, types.NoneType):
        return NoneToPyNone(type, mapper, artifact, path)
    elif isinstance(type, types.UnknownType):
        return UnknownToPyUnknown(type, mapper, artifact, path)
    elif isinstance(type, types.RefType):
        return PyRefToRef(type, mapper, artifact, path)
    else:
        for m in _additional_mappers:
            if isinstance(type, m.type_class):
                return m.from_mapper(type, mapper, artifact, path)
        return DefaultFromPy(type, mapper, artifact, path)


mappers_python.map_to_python = mappers.make_mapper(map_to_python_)
mappers_python.map_from_python = mappers.make_mapper(map_from_python_)



# Language feature: Automatic function calling
#
# A Weave Node is a Function. If there are no variables in the Node's DAG,
# the node can be executed.
#
# If you have a Node[Function[..., int]], you may pass it in positions that
# expect int. We automatically insert an .execute() op when this type of
# call is detected.
#
# TODO: resolve ambiguities:
#   - should it work recursively? (Node[Function[..., Function[..., int]]])
#   - does this cause invalid type/name overlap and collisions?

# There is also a compile pass for this in compile.py

import typing

from . import weave_types as types
from . import op_args


def update_input_types(
    op_input_type: typing.Optional[op_args.OpArgs],
    actual_input_types: dict[str, types.Type],
) -> dict[str, types.Type]:
    if not isinstance(op_input_type, op_args.OpNamedArgs):
        return actual_input_types
    expected_input_types = op_input_type.arg_types
    result: dict[str, types.Type] = {}
    try:
        for k, t in actual_input_types.items():
            expected_input_type = expected_input_types[k]
            if (
                isinstance(t, types.Function)
                and not callable(expected_input_type)
                and not isinstance(expected_input_type, types.Function)
            ):
                result[k] = t.output_type
            else:
                result[k] = t
    except KeyError:
        return actual_input_types
    return result



from weave import weave_types
from ..ops_arrow.arrow import ArrowWeaveListType
from ..core_types import StreamTableType
from ..api import op
from .project_ops import project
from .. import compile


def _get_history_node(steam_table: StreamTableType):
    return (
        project(steam_table.entity_name, steam_table.project_name)
        .run(steam_table.table_name)
        .history3()
    )


@op()
def _rows_type(steam_table: StreamTableType) -> weave_types.Type:
    with compile.enable_compile():
        return compile.compile([_get_history_node(steam_table)])[0].type


@op(
    name="stream_table-rows",
    output_type=ArrowWeaveListType(weave_types.TypedDict({})),
    refine_output_type=_rows_type,
    returns_expansion_node=True,
)
def rows(steam_table: StreamTableType):
    return _get_history_node(steam_table)



from .number import *
from .string import *
from .boolean import *
from .date import *
from .obj import *
from .dict import *
from .list_join import *
from .list_ops import *
from .convert_ops import *
from .arraylist_ops import *



# The idea is that we provide the same interface from reading the Slack Api, or a Slack data export
import typing


class SlackApi(typing.Protocol):
    def channel_names(self):
        ...

    def channel_export_size(self, channel_name: str):
        ...

    # Returns a dict in the slack export format. TODO
    def channel_messages(self, channel_name: str):
        ...



# Very much WIP. An implementation of the read side of the next gen
# W&B runs/table engine.

import itertools
import random
import string
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.parquet as pc
import pyarrow.dataset as ds
import pyarrow.feather as pf
import pyarrow.compute as pc

import typing
import weave
from weave.ops_domain import wb_domain_types

from .. import engine_trace
from ..ops_arrow.arrow import arrow_schema_to_weave_type
from ..ops_arrow.list_ import ArrowWeaveListType


class EmptyTypedDict(typing.TypedDict):
    pass


class Run2(typing.TypedDict):
    id: str
    config: EmptyTypedDict
    summary: EmptyTypedDict


def get_runs_ds_columns(
    config_cols: list[str], summary_cols: list[str]
) -> dict[str, pc.Expression]:
    # TODO: only handles the first level of nesting right now
    cols = {
        "id": ds.field("id"),
    }
    if not config_cols:
        # TODO: can't default to key0, but pyarrow doeesn't like empty values.
        config_cols = ["key0"]
    if not summary_cols:
        # TODO: can't default to key0, but pyarrow doeesn't like empty values.
        summary_cols = ["key0"]
    if config_cols:
        cols["config"] = pc.make_struct(
            *(ds.field("config", k) for k in config_cols), field_names=config_cols
        )
    if summary_cols:
        cols["summary"] = pc.make_struct(
            *(ds.field("summary", k) for k in summary_cols), field_names=summary_cols
        )
    return cols


def read_runs(
    project_name: str, columns=None, limit=None
) -> weave.ops.ArrowWeaveList[Run2]:
    tracer = engine_trace.tracer()

    # Uncomment to use feather format instead
    # with tracer.trace("read_arch"):
    #     # This is the slow part!
    #     arch = ds.dataset(f"/tmp/runs.{project_name}.feather", format='feather').to_table(columns=columns)
    # with tracer.trace("read_live"):
    #     live = ds.dataset(f"/tmp/runs.{project_name}.live.feather", format='feather').to_table(
    #         columns=columns
    #     )
    with tracer.trace("read_arch"):
        # This is the slow part!
        arch = ds.dataset(f"/tmp/runs.{project_name}.parquet").to_table(columns=columns)
    with tracer.trace("read_live"):
        live = ds.dataset(f"/tmp/runs.{project_name}.live.parquet").to_table(
            columns=columns
        )
    import duckdb

    with tracer.trace("ddb_connect"):
        con = duckdb.connect()
    if limit is not None:
        limit = f"LIMIT {limit}"
    else:
        limit = ""
    with tracer.trace("ddb_join"):
        # 'live' and 'arch' refer to tables variables in the local scope!
        joined = con.execute(
            """
            SELECT COALESCE(ls.id, p.id) as id,
                COALESCE(ls.config, p.config) as config,
                COALESCE(ls.summary, p.summary) as summary,
            FROM 'live' as ls
            FULL OUTER JOIN 'arch' as p
            ON ls.id=p.id
            %s
            """
            % limit
        )
    with tracer.trace("ddb_arrow"):
        joined_table = joined.arrow()

    with tracer.trace("to_weave"):
        # TODO: Should this pass an artifact to `weave.ops.ArrowWeaveList`?
        res: weave.ops.ArrowWeaveList[Run2] = weave.ops.ArrowWeaveList(joined_table)
    return res


@weave.op(render_info={"type": "function"}, hidden=True)
def refine_runs2_with_columns_type(
    project: wb_domain_types.Project, config_cols: list[str], summary_cols: list[str]
) -> weave.types.Type:
    # TODO: only reading type from archived runs right now. There is a bunch of logic we need
    # to get this right
    schema = ds.dataset(f"/tmp/runs.{project.gql['name']}.parquet").schema
    full_run_type = typing.cast(
        weave.types.TypedDict, arrow_schema_to_weave_type(schema)
    )
    full_config_type = typing.cast(
        weave.types.TypedDict, full_run_type.property_types["config"]
    )
    full_summary_type = typing.cast(
        weave.types.TypedDict, full_run_type.property_types["config"]
    )
    partial_run_type = weave.types.TypedDict(
        {
            "id": full_run_type.property_types["id"],
            "config": weave.types.TypedDict(
                {k: full_config_type.property_types[k] for k in config_cols}
            ),
            "summary": weave.types.TypedDict(
                {k: full_summary_type.property_types[k] for k in summary_cols}
            ),
        }
    )
    return ArrowWeaveListType(partial_run_type)


@weave.op(
    name="project-runs2_with_columns",
    refine_output_type=refine_runs2_with_columns_type,
)
def runs2_with_columns(
    project: wb_domain_types.Project, config_cols: list[str], summary_cols: list[str]
) -> weave.ops.ArrowWeaveList[Run2]:
    return read_runs(
        project.gql["name"], columns=get_runs_ds_columns(config_cols, summary_cols)
    )


@weave.op(render_info={"type": "function"}, hidden=True)
def refine_runs2_type(project: wb_domain_types.Project) -> weave.types.Type:
    schema = ds.dataset(f"/tmp/runs.{project.gql['name']}.parquet").schema
    return ArrowWeaveListType(arrow_schema_to_weave_type(schema))


@weave.op(name="project-runs2", refine_output_type=refine_runs2_type)
def runs2(project: wb_domain_types.Project) -> weave.ops.ArrowWeaveList[Run2]:
    return read_runs(project.gql["name"])


# Defines a panel with all columns shown for the runs
@weave.op()
def render_table_runs2(runs: weave.Node[list[Run2]]) -> weave.panels.Table:
    run_type = runs.type.object_type  # type: ignore
    config_type = run_type.property_types["config"]
    summary_type = run_type.property_types["summary"]
    columns = []

    def col_key(field, key):
        return lambda run: run[field][key]

    for key in config_type.property_types:
        columns.append(col_key("config", key))
    for key in summary_type.property_types:
        columns.append(col_key("summary", key))
    return weave.panels.Table(runs, columns=columns)


# Data generation for testing
def random_string(n):
    return "".join(random.choices(string.ascii_lowercase, k=n))


def make_runs(n_runs, n_summary, n_config, n_config_value_chars):
    config_choices = [
        "".join(c)
        for c in itertools.combinations(string.ascii_lowercase, n_config_value_chars)
    ]
    ids = [random_string(8) for i in range(n_runs)]
    config_arrays = [np.random.choice(config_choices, n_runs) for i in range(n_config)]
    config_keys = [f"key{i}" for i in range(n_config)]
    config = pa.StructArray.from_arrays(config_arrays, config_keys)
    summary_arrays = [np.random.random(n_runs) for i in range(n_summary)]
    summary_keys = [f"key{i}" for i in range(n_summary)]
    summary = pa.StructArray.from_arrays(summary_arrays, summary_keys)
    return pa.Table.from_arrays([ids, config, summary], ["id", "config", "summary"])


def make_runs2_tables(n_runs, n_summary, n_config, n_config_value_chars, live_set_size):
    runs = make_runs(n_runs, n_summary, n_config, n_config_value_chars)
    arch_runs = runs[: len(runs) - live_set_size]
    liveset_runs = runs[len(runs) - live_set_size :]
    proj_name = f"weavetest-{n_runs}-{n_summary}-{n_config}-{live_set_size}"
    pq.write_table(arch_runs, f"/tmp/runs.{proj_name}.parquet")
    pf.write_feather(arch_runs, f"/tmp/runs.{proj_name}.feather")
    pq.write_table(liveset_runs, f"/tmp/runs.{proj_name}.live.parquet")
    pf.write_feather(liveset_runs, f"/tmp/runs.{proj_name}.live.feather")



import numpy as np
from . import weave_types as types
from . import errors


class NumpyArraySaver:
    def __init__(self, artifact, name):
        self._artifact = artifact
        self._name = name
        self._arrays = []

    def add(self, arr):
        self._arrays.append(arr)
        return len(self._arrays) - 1

    def close(self):
        arr = np.array(self._arrays)
        with self._artifact.new_file(f"{self._name}.npz", binary=True) as f:
            np.savez_compressed(f, arr=arr)


class NumpyArrayLoader:
    def __init__(self, artifact, name):
        self._artifact = artifact
        self._name = name
        with artifact.open(f"{name}.npz", binary=True) as f:
            self._arr = np.load(f)["arr"]

    def get(self, index):
        return self._arr[index]


# TODO: this doesn't match how extra works for list types...
class NumpyArrayType(types.Type):
    instance_classes = np.ndarray
    name = "WeaveNDArray"

    def __init__(self, dtype="x", shape=(0,)):
        self.dtype = dtype
        self.shape = shape

    @classmethod
    def type_of_instance(cls, obj):
        return cls(obj.dtype, obj.shape)

    def _to_dict(self):
        return {"dtype": str(self.dtype), "shape": self.shape}

    @classmethod
    def from_dict(cls, d):
        return cls(np.dtype(d.get("dtype", "object")), d["shape"])

    def _assign_type_inner(self, next_type):
        if not isinstance(next_type, NumpyArrayType):
            return False
        if (
            self.dtype != next_type.dtype
            and next_type.dtype != np.dtype("object")  # object is like "any"
        ) or tuple(self.shape) != tuple(next_type.shape):
            return False
        return True

    def save_instance(self, obj, artifact, name):
        handler = artifact.get_path_handler(name, handler_constructor=NumpyArraySaver)
        index = handler.add(obj)
        return [str(index)]

    @classmethod
    def load_instance(cls, artifact, name, extra=None):
        if extra is None or not isinstance(extra, list):
            raise errors.WeaveInternalError(
                f"Received unexpect 'extra' param - {extra}. Expected a singleton list of integer."
            )
        extra = extra[0]
        index = int(extra)
        handler = artifact.get_path_handler(name, handler_constructor=NumpyArrayLoader)
        return handler.get(index)

    def __str__(self):
        return "<NumpyArrayType %s %s>" % (self.dtype, self.shape)



import dataclasses
import json
import pyarrow as pa
import numpy as np
from pyarrow import compute as pc
import pyarrow.parquet as pq
import typing
import textwrap

py_type = type

from .. import mappers_python
from .. import weave_types as types
from .. import errors
from .. import artifact_fs


def arrow_type_to_weave_type(pa_type: pa.DataType) -> types.Type:
    if pa_type == pa.null():
        return types.NoneType()
    elif pa.types.is_dictionary(pa_type):
        return arrow_type_to_weave_type(pa_type.value_type)
    elif pa_type == pa.string():
        return types.String()
    elif pa_type == pa.large_string():
        return types.String()
    elif pa_type == pa.int64() or pa_type == pa.int32():
        return types.Int()
    elif pa_type in [pa.float64(), pa.float32(), pa.float16()]:
        return types.Float()
    elif pa_type == pa.bool_():
        return types.Boolean()
    elif pa.types.is_temporal(pa_type):
        return types.Timestamp()
    elif pa.types.is_binary(pa_type):
        return types.Bytes()
    elif pa.types.is_list(pa_type):
        return types.List(arrow_field_to_weave_type(pa_type.value_field))
    elif pa.types.is_struct(pa_type):
        return types.TypedDict({f.name: arrow_field_to_weave_type(f) for f in pa_type})
    elif pa.types.is_union(pa_type):
        return types.UnionType(*[arrow_type_to_weave_type(f.type) for f in pa_type])
    raise errors.WeaveTypeError(
        "Type conversion not implemented for arrow type: %s" % pa_type
    )


def arrow_field_to_weave_type(pa_field: pa.Field) -> types.Type:
    t = arrow_type_to_weave_type(pa_field.type)
    if pa_field.nullable:
        return types.optional(t)
    return t


def arrow_schema_to_weave_type(schema) -> types.Type:
    return types.TypedDict({f.name: arrow_field_to_weave_type(f) for f in schema})


@dataclasses.dataclass(frozen=True)
class ArrowArrayType(types.Type):
    instance_classes = [pa.ChunkedArray, pa.ExtensionArray, pa.Array]
    name = "ArrowArray"

    object_type: types.Type = types.Any()
    UNION_PREFIX = "__weave_union"
    UNION_TO_STRUCT_TYPE_CODE_COLNAME = "__weave_type_code"
    EMPTY_STRUCT_DUMMY_FIELD_NAME = "__weave_empty_struct"

    @classmethod
    def type_of_instance(cls, obj: pa.Array):
        return cls(arrow_type_to_weave_type(obj.type))

    def save_instance(self, obj, artifact, name):
        table = pa.table({"arr": obj})
        with artifact.new_file(f"{name}.parquet", binary=True) as f:
            pq.write_table(table, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.parquet", binary=True) as f:
            deserialized = pq.read_table(f)
            deserialized = deserialized["arr"].combine_chunks()
            return deserialized


@dataclasses.dataclass(frozen=True)
class ArrowTableType(types.Type):
    instance_classes = pa.Table
    name = "ArrowTable"

    object_type: types.Type = types.Any()

    @classmethod
    def type_of_instance(cls, obj: pa.Table):
        return cls(arrow_schema_to_weave_type(obj.schema))

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.parquet", binary=True) as f:
            pq.write_table(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.parquet", binary=True) as f:
            return pq.read_table(f)


# This function is evil and breaks performance because of the dictionary_decode
# call. Don't use it. TODO: remove all calls
def arrow_as_array(obj) -> pa.Array:
    # assumes obj is table or array
    if isinstance(obj, pa.Table):
        return pa.StructArray.from_arrays(
            # TODO: we shouldn't need to combine chunks, we can produce this in the
            # original chunked form for zero copy
            [c.combine_chunks() for c in obj.columns],
            names=obj.column_names,
        )
    elif isinstance(obj, pa.ChunkedArray):
        return obj.combine_chunks()
    elif isinstance(obj, pa.DictionaryArray):
        return arrow_as_array(obj.dictionary_decode())
    elif not isinstance(obj, pa.Array):
        raise TypeError("Expected pyarrow Array or Table, got %s" % type(obj))
    return obj


def offsets_starting_at_zero(arr: pa.ListArray) -> pa.IntegerArray:
    """
    We often have code to operate on the elements of a list array. That code
    will do something like:
    ```
        offsets = arr.offsets
        elements = arr.flatten()
        transformed = do_something(elements)
        result = pa.ListArray.from_arrays(offsets, transformed)
    ```
    However, sometimes for performance reasons, the offsets do not start at 0.
    For example, when the ListArray is actually a slice of a larger underlying
    buffer. We should use this helper function anytime we are doing something
    like the above example.
    """
    raw_offsets = arr.offsets
    if len(raw_offsets) == 0:
        return raw_offsets
    first_value = raw_offsets[0]
    if first_value == 0:
        return raw_offsets
    else:
        return pc.subtract(raw_offsets, first_value)


@dataclasses.dataclass(frozen=True)
class ArrowWeaveListType(types.Type):
    _base_type = types.List
    name = "ArrowWeaveList"

    object_type: types.Type = types.Any()

    @classmethod
    def type_of_instance(cls, obj):
        return cls(obj.object_type)

    def save_instance(self, obj, artifact, name):
        # rewriting refs is disabled after a refactor. We need this to work
        # for media and caching with local artifacts to work. But we don't need
        # it for W&B production for weave0 use cases.
        # TODO: fix this

        # If we are saving to the same artifact as we were written to,
        # then we don't need to rewrite any references.

        # if obj._artifact == artifact:
        #     arrow_data = obj._arrow_data
        # else:
        #     # super().save_instance(obj, artifact, name)
        #     # return
        #     arrow_data = rewrite_weavelist_refs(
        #         obj._arrow_data, obj.object_type, obj._artifact, artifact
        #     )
        if not obj._artifact == artifact:
            # somehow we still need this block of code to get the test passing,
            # even though we don't use its result.
            rewrite_weavelist_refs(
                obj._arrow_data, obj.object_type, obj._artifact, artifact
            )

        from . import convert

        parquet_friendly = convert.to_parquet_friendly(obj)
        table = pa.table({"arr": parquet_friendly._arrow_data})
        with artifact.new_file(f"{name}.ArrowWeaveList.parquet", binary=True) as f:
            pq.write_table(table, f)
        with artifact.new_file(f"{name}.ArrowWeaveList.type.json") as f:
            json.dump(obj.object_type.to_dict(), f)

    def load_instance(
        self, artifact: artifact_fs.FilesystemArtifact, name: str, extra=None
    ):
        with artifact.open(f"{name}.ArrowWeaveList.parquet", binary=True) as f:
            table = pq.read_table(f)
        arr = table["arr"].combine_chunks()
        with artifact.open(f"{name}.ArrowWeaveList.type.json") as f:
            object_type = json.load(f)
            object_type = types.TypeRegistry.type_from_dict(object_type)
        from . import list_

        with list_.unsafe_awl_construction("load_from_parquet"):
            l = self.instance_class(arr, object_type=object_type, artifact=artifact)  # type: ignore
            from . import convert

            res = convert.from_parquet_friendly(l)
        res.validate()
        return res


def rewrite_weavelist_refs(arrow_data, object_type, source_artifact, target_artifact):
    if _object_type_has_props(object_type):
        prop_types = _object_type_prop_types(object_type)

        # handle empty struct case - the case where the struct has no fields
        if len(prop_types) == 0:
            return arrow_data

        if isinstance(arrow_data, pa.Table):
            arrays = {}
            for col_name, col_type in prop_types.items():
                column = arrow_data[col_name]
                arrays[col_name] = rewrite_weavelist_refs(
                    column, col_type, source_artifact, target_artifact
                )
            return pa.table(arrays)
        elif isinstance(arrow_data, pa.ChunkedArray):
            arrays = {}
            unchunked = arrow_data.combine_chunks()
            for col_name, col_type in prop_types.items():
                column = unchunked.field(col_name)
                arrays[col_name] = rewrite_weavelist_refs(
                    column, col_type, source_artifact, target_artifact
                )
            return pa.StructArray.from_arrays(
                arrays.values(), names=arrays.keys(), mask=pa.compute.is_null(unchunked)
            )
        elif isinstance(arrow_data, pa.StructArray):
            arrays = {}
            for col_name, col_type in prop_types.items():
                column = arrow_data.field(col_name)
                arrays[col_name] = rewrite_weavelist_refs(
                    column, col_type, source_artifact, target_artifact
                )
            return pa.StructArray.from_arrays(
                arrays.values(),
                names=arrays.keys(),
                mask=pa.compute.is_null(arrow_data),
            )
        elif isinstance(arrow_data, pa.NullArray):
            return arrow_data
        else:
            raise errors.WeaveTypeError('Unhandled type "%s"' % type(arrow_data))
    elif isinstance(object_type, types.UnionType):
        non_none_members = [
            m for m in object_type.members if not isinstance(m, types.NoneType)
        ]
        nullable = len(non_none_members) < len(object_type.members)
        if len(non_none_members) > 1:
            arrow_data = arrow_as_array(arrow_data)
            if not isinstance(arrow_data.type, pa.UnionType):
                raise errors.WeaveTypeError(
                    "Expected UnionType, got %s" % type(arrow_data.type)
                )
            arrays = []
            for i, _ in enumerate(arrow_data.type):
                rewritten = rewrite_weavelist_refs(
                    arrow_data.field(i),
                    non_none_members[i]
                    if not nullable
                    else types.UnionType(types.NoneType(), non_none_members[i]),
                    source_artifact,
                    target_artifact,
                )
                arrays.append(rewritten)
            return pa.UnionArray.from_dense(
                arrow_data.type_codes, arrow_data.offsets, arrays
            )
        return rewrite_weavelist_refs(
            arrow_data, types.non_none(object_type), source_artifact, target_artifact
        )
    elif _object_type_is_basic(object_type):
        return arrow_data
    elif isinstance(object_type, (types.List, ArrowWeaveListType)):
        data = arrow_data
        if isinstance(arrow_data, pa.ChunkedArray):
            data = arrow_data.combine_chunks()
        return pa.ListArray.from_arrays(
            offsets_starting_at_zero(data),
            rewrite_weavelist_refs(
                data.flatten(),
                object_type.object_type,
                source_artifact,
                target_artifact,
            ),
            mask=pa.compute.is_null(data),
        )
    else:
        # We have a column of refs
        new_refs = []
        for ref_str in arrow_data:
            ref_str = ref_str.as_py()
            new_refs.append(
                _rewrite_ref_for_save(
                    ref_str, object_type, source_artifact, target_artifact
                )
                if ref_str is not None
                else None
            )
        return pa.array(new_refs)


def _object_type_has_props(object_type):
    from ..language_features.tagging import tagged_value_type

    return (
        isinstance(object_type, types.TypedDict)
        or isinstance(object_type, types.ObjectType)
        or isinstance(object_type, tagged_value_type.TaggedValueType)
    )


def _object_type_prop_types(object_type):
    from ..language_features.tagging import tagged_value_type

    if isinstance(object_type, tagged_value_type.TaggedValueType):
        return {
            "_tag": object_type.tag,
            "_value": object_type.value,
        }
    prop_types = object_type.property_types
    if callable(prop_types):
        prop_types = prop_types()
    return prop_types


def _object_type_is_basic(object_type):
    if isinstance(object_type, types.Const):
        object_type = object_type.val_type
    return isinstance(object_type, types.BasicType) or (
        isinstance(object_type, types.Timestamp)
    )


def _rewrite_ref_for_save(entry: str, object_type, source_artifact, target_artifact):
    # TODO: This should be simpler. We take a ref string, which may be absolute or
    # relative, and want to ensure the object is referencable from the target_artifact.
    # This function should be written in those terms.
    if ":" in entry:
        # Already a URI
        return entry

    if isinstance(source_artifact, artifact_fs.FilesystemArtifact):
        # Our source is an artifact, construct a full URI.
        return str(source_artifact.ref_from_local_str(entry, object_type).uri)

    # Source is ObjLookupMem, save to the artifact
    return target_artifact.set(
        entry, object_type, source_artifact.get(entry, object_type)
    ).local_ref_str()


def pretty_print_arrow_type(t: typing.Union[pa.Schema, pa.DataType, pa.Field]) -> str:
    if isinstance(t, pa.Schema):
        return "Schema:\n" + textwrap.indent(
            "\n".join(pretty_print_arrow_type(f) for f in t), "  "
        )
    elif isinstance(t, pa.Field):
        return f"{t.name}: {'nullable' if t.nullable else 'non-null'} {pretty_print_arrow_type(t.type)}"

    if isinstance(t, pa.StructType):
        return "Struct:\n" + textwrap.indent(
            "\n".join(pretty_print_arrow_type(f) for f in t), "  "
        )

    elif isinstance(t, pa.ListType):
        return f"List\n" + textwrap.indent(pretty_print_arrow_type(t.value_type), "  ")

    elif isinstance(t, pa.UnionType):
        return "Union:\n" + textwrap.indent(
            "\n".join(pretty_print_arrow_type(f) for f in t), "  "
        )

    return str(t)


def union_is_null(arr: pa.Array):
    # pyarrow doesn't support is_null on unions. We do it ourselves
    # by checking each child entry for whether its null. This code
    # is vectorized so still fast.
    merged = pa.nulls(len(arr), pa.bool_())
    for type_code in range(len(arr.type)):
        field = pa.compute.is_null(arr.field(type_code))
        if len(field) == 0:
            continue
        mask = pa.compute.equal(arr.type_codes, type_code)
        indexes = pa.compute.multiply(mask.cast(pa.int8()), arr.offsets)
        values = field.take(indexes)
        merged = pa.compute.if_else(mask, values, merged)
    return merged


def safe_is_null(arr: pa.Array):
    # is_null that also handles unions correctly since pyarrow doesn't.
    if isinstance(arr.type, pa.UnionType):
        return union_is_null(arr)
    else:
        return pa.compute.is_null(arr)


def safe_coalesce(*arrs: pa.Array):
    if not arrs:
        raise ValueError("coalesce requires at least one argument")
    # pyarrows coalesce doesn't handle unions. This does!
    if len(arrs[0]) == 0:
        return arrs[0]
    result = arrs[0]
    for arr in arrs[1:]:
        result = pa.compute.if_else(safe_is_null(result), arr, result)
    return result


def arrow_zip(*arrs: pa.Array) -> pa.Array:
    n_arrs = len(arrs)
    output_len = min(len(a) for a in arrs)
    array_indexes = np.tile(np.arange(n_arrs, dtype="int64"), output_len)
    item_indexes = np.floor(np.arange(0, output_len, 1.0 / n_arrs)).astype("int64")
    indexes = item_indexes + array_indexes * output_len
    concatted = pa.concat_arrays(arrs)
    interleaved = concatted.take(indexes)
    offsets = np.arange(0, len(interleaved) + len(arrs), len(arrs), dtype="int64")
    return pa.ListArray.from_arrays(offsets, interleaved)



# This is deprecated in favor of the new wandb_api.py module.
# TODO: remove uses of this and delete.

from wandb.apis import public
from wandb.sdk.internal.internal_api import _thread_local_api_settings
import typing

from . import errors


def wandb_public_api() -> public.Api:
    return public.Api(timeout=30)


def assert_wandb_authenticated() -> None:
    authenticated = wandb_public_api().api_key is not None
    if not authenticated:
        raise errors.WeaveWandbAuthenticationException(
            f"Unable to log data to W&B. Please authenticate by setting WANDB_API_KEY or running `wandb init`."
        )


def wandb_gql_query(
    query_str: str, variables: dict[str, typing.Any] = {}
) -> typing.Any:
    return wandb_public_api().client.execute(
        public.gql(query_str),
        variable_values=variables,
    )


def set_wandb_thread_local_api_settings(
    api_key: typing.Optional[str],
    cookies: typing.Optional[typing.Dict],
    headers: typing.Optional[typing.Dict],
) -> None:
    _thread_local_api_settings.api_key = api_key
    _thread_local_api_settings.cookies = cookies
    _thread_local_api_settings.headers = headers


class WandbThreadLocalApiSettings(typing.TypedDict):
    api_key: typing.Optional[str]
    cookies: typing.Optional[dict]
    headers: typing.Optional[dict]


def reset_wandb_thread_local_api_settings() -> None:
    _thread_local_api_settings.api_key = None
    _thread_local_api_settings.cookies = None
    _thread_local_api_settings.headers = None



# Disable warning spew.
import warnings

warnings.filterwarnings("ignore", message=".*The 'nopython' keyword.*")

import shap
import random
import typing
import xgboost
import pickle
import numpy as np

import matplotlib.pyplot as plt
import weave

from .. import huggingface as hf
from .. import xgboost as weave_xgb


@weave.op(
    input_type={
        "df": weave.types.TypedDict({}),
    },
    output_type=weave.types.TypedDict({"X": weave.types.Any(), "y": weave.types.Any()}),
    hidden=True,
)
def split_labels(df, label_col: str):
    X = df.drop(label_col, axis=1)
    y = df[label_col]
    return {"X": X, "y": y}


@weave.type()
class ShapValues:
    values: np.ndarray

    @weave.op(
        output_type=weave.types.FileType(weave.types.Const(weave.types.String(), "png"))  # type: ignore
    )
    def summary_plot(self):
        shap.summary_plot(self.values, show=False)
        path = "/tmp/shap-%s.png" % random.randrange(0, 1000000)
        plt.savefig(path)
        plt.close()
        return weave.ops.LocalFile(path)


class ShapExplanationType(weave.types.Type):
    instance_classes = shap.Explanation

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pickle", binary=True) as f:
            pickle.dump(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pickle", binary=True) as f:
            return pickle.load(f)


@weave.op()
def shap_explain_tree(self: xgboost.core.Booster, data: typing.Any) -> ShapValues:
    explainer = shap.TreeExplainer(self)
    shap_values = explainer.shap_values(data)
    return ShapValues(shap_values)


@weave.op()
def shap_explain(
    pipeline_output: hf.FullTextClassificationPipelineOutput,
) -> shap.Explanation:
    # TODO: shap has some options, like computing logits is better in some cases?
    # TODO: does shap work for all task styles?
    pipeline = weave.use(pipeline_output._model.pipeline())
    explainer = shap.Explainer(pipeline)
    return explainer([pipeline_output.model_input])


@weave.op()
def shap_plot_text(shap_values: shap.Explanation) -> weave.ops.Html:
    html = shap.plots.text(shap_values, display=False)
    return weave.ops.Html(html)


@weave.type()
class ShapPlotText(weave.Panel):
    id = "ShapPlotText"
    input_node: weave.Node[shap.Explanation]

    @weave.op()
    def render(self) -> weave.panels.PanelHtml:
        return weave.panels.PanelHtml(shap_plot_text(self.input_node))



import math

from ..api import op, use
from ..weave_types import Function, NumberBinType
from .. import weave_types as types
from .. import graph
from .dict import dict_
from . import date

from ..weave_internal import define_fn, call_fn, make_const_node


@op(
    input_type={"step": types.Number()},
    output_type=Function(
        input_types={"row": types.Number()},
        output_type=NumberBinType,  # type: ignore
    ),
)
def number_bins_fixed(step):
    def body(row):
        if step <= 0:
            raise ValueError("Step must be greater than zero.")
        mult = make_const_node(types.Number(), 1.0 / step)
        start_node = (row * mult).floor() / mult
        return dict_(start=start_node, stop=start_node + step)

    # this should be vectorized by expansion
    return define_fn({"row": types.Number()}, body).val


@op(
    name="numbers-binsequal",
    input_type={
        "arr": types.List(types.optional(types.Number())),
        "bins": types.Number(),
    },
    output_type=Function(
        input_types={"row": types.Number()}, output_type=NumberBinType  # type: ignore
    ),
    render_info={"type": "function"},
)
def numbers_bins_equal(arr, bins):
    arr_min = min(arr) if len(arr) > 0 else 0
    arr_max = max(arr) if len(arr) > 0 else 0
    step = (arr_max - arr_min) / bins
    return use(number_bins_fixed(step))


@op(
    name="number-bin",
    input_type={
        "in_": types.Number(),
        "bin_fn": types.optional(
            Function(
                input_types={"row": types.Number()}, output_type=NumberBinType  # type: ignore
            )
        ),
    },
    output_type=NumberBinType,  # type: ignore
)
def number_bin(in_, bin_fn):
    if not isinstance(bin_fn, graph.Node) and bin_fn == None:
        return None
    return use(call_fn(bin_fn, {"row": make_const_node(types.Number(), in_)}))



import json

from ..compile_domain import wb_gql_op_plugin
from ..api import op
from .. import errors
from .. import weave_types as types
from . import wb_domain_types as wdt
from .wandb_domain_gql import (
    _make_alias,
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
)

import typing
from . import wb_util
from urllib.parse import quote
from .. import artifact_fs
from .. import artifact_wandb


static_art_file_gql = """
            commitHash
            artifactSequence {
                id
                name
                defaultArtifactType {
                    id
                    name
                    project {
                        id
                        name
                        entity {
                            id
                            name
                        }
                    }
                }
            }
        """

# Section 1/6: Tag Getters
# None


def _safe_nested_pick(obj: typing.Optional[dict], path: list[str]) -> typing.Any:
    if len(path) == 0:
        return obj
    if not isinstance(obj, dict):
        return None
    return _safe_nested_pick(obj.get(path[0]), path[1:])


# Section 2/6: Root Ops
@op(
    name="root-artifactVersionGQLResolver",
    input_type={
        "gql_result": types.TypedDict({}),
        "entityName": types.String(),
        "projectName": types.String(),
        "artifactTypeName": types.String(),
        "artifactVersionName": types.String(),
    },
    output_type=types.optional(wdt.ArtifactVersionType),
    hidden=True,
)
def root_artifact_version_gql_resolver(
    gql_result, entityName, projectName, artifactTypeName, artifactVersionName
):
    project_alias = _make_alias(entityName, projectName, prefix="project")
    artifact_type_alias = _make_alias(artifactTypeName, prefix="artifactType")
    artifact_alias = _make_alias(artifactVersionName, prefix="artifact")

    # This nested property access is required because if an particular
    # GQL does not exist, then it returns None for that key. So, if
    # the user requests something that does not exist, then we want
    # to safely return None in such a case
    artifact_alias_data = _safe_nested_pick(
        gql_result, [project_alias, artifact_type_alias, artifact_alias]
    )
    if artifact_alias_data is not None:
        return wdt.ArtifactType.from_gql(artifact_alias_data)
    return None


def _root_artifact_version_plugin(inputs, inner):
    project_alias = _make_alias(
        inputs.raw["entityName"], inputs.raw["projectName"], prefix="project"
    )
    artifact_type_alias = _make_alias(
        inputs.raw["artifactTypeName"], prefix="artifactType"
    )
    artifact_alias = _make_alias(inputs.raw["artifactVersionName"], prefix="artifact")
    return f"""
        {project_alias}: project(entityName: {inputs['entityName']}, name:{inputs['projectName']}){{
            id
            {artifact_type_alias}: artifactType(name: {inputs['artifactTypeName']}){{
                id
                {artifact_alias}: artifact(name: {inputs['artifactVersionName']}){{
                    {wdt.ArtifactVersion.REQUIRED_FRAGMENT}
                    {inner}
                }}
            }}
        }}
    """


@op(
    name="root-artifactVersion",
    input_type={
        "entityName": types.String(),
        "projectName": types.String(),
        "artifactTypeName": types.String(),
        "artifactVersionName": types.String(),
    },
    output_type=wdt.ArtifactVersionType,
    plugins=wb_gql_op_plugin(
        _root_artifact_version_plugin,
        is_root=True,
        root_resolver=root_artifact_version_gql_resolver,
    ),
)
def root_artifact_version(
    entityName, projectName, artifactTypeName, artifactVersionName
):
    raise errors.WeaveGQLCompileError(
        "root-artifactVersion should not be executed directly. If you see this error, it is a bug in the Weave compiler."
    )


# Section 3/6: Attribute Getters
gql_prop_op("artifactVersion-id", wdt.ArtifactVersionType, "id", types.String())
gql_prop_op("artifactVersion-digest", wdt.ArtifactVersionType, "digest", types.String())
gql_prop_op(
    "artifactVersion-hash", wdt.ArtifactVersionType, "commitHash", types.String()
)
gql_prop_op("artifactVersion-size", wdt.ArtifactVersionType, "size", types.Int())
gql_prop_op(
    "artifactVersion-description", wdt.ArtifactVersionType, "description", types.Int()
)
gql_prop_op(
    "artifactVersion-createdAt", wdt.ArtifactVersionType, "createdAt", types.Timestamp()
)

gql_prop_op(
    "artifactVersion-versionId", wdt.ArtifactVersionType, "versionIndex", types.Number()
)

gql_prop_op(
    "artifactVersion-referenceCount",
    wdt.ArtifactVersionType,
    "usedCount",
    types.Number(),
)


@op(plugins=wb_gql_op_plugin(lambda inputs, inner: "metadata"), hidden=True)
def refine_metadata(
    artifactVersion: wdt.ArtifactVersion,
) -> types.Type:
    return wb_util.process_run_dict_type(
        json.loads(artifactVersion.gql["metadata"] or "{}")
    )


@op(
    name="artifactVersion-metadata",
    refine_output_type=refine_metadata,
    plugins=wb_gql_op_plugin(lambda inputs, inner: "metadata"),
)
def metadata(
    artifactVersion: wdt.ArtifactVersion,
) -> dict[str, typing.Any]:
    return wb_util.process_run_dict_obj(
        json.loads(artifactVersion.gql["metadata"] or "{}")
    )


# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "artifactVersion-aliases",
    wdt.ArtifactVersionType,
    "aliases",
    wdt.ArtifactAliasType,
    is_many=True,
)

gql_direct_edge_op(
    "artifactVersion-artifactType",
    wdt.ArtifactVersionType,
    "artifactType",
    wdt.ArtifactTypeType,
)

gql_direct_edge_op(
    "artifactVersion-artifactSequence",
    wdt.ArtifactVersionType,
    "artifactSequence",
    wdt.ArtifactCollectionType,
)


@op(
    name="artifactVersion-createdBy",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
        createdBy {{
            __typename
            ... on Run {{
                {wdt.Run.REQUIRED_FRAGMENT}
                {inner}
            }}
        }}
        """
    ),
)
def artifact_version_created_by(
    artifactVersion: wdt.ArtifactVersion,
) -> typing.Optional[wdt.Run]:
    if artifactVersion.gql["createdBy"]["__typename"] == "Run":
        return wdt.Run.from_gql(artifactVersion.gql["createdBy"])
    return None


@op(
    name="artifactVersion-createdByUser",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
        createdBy {{
            __typename
            ... on User {{
                {wdt.User.REQUIRED_FRAGMENT}
                {inner}
            }}
        }}
        """
    ),
)
def artifact_version_created_by_user(
    artifactVersion: wdt.ArtifactVersion,
) -> typing.Optional[wdt.User]:
    if artifactVersion.gql["createdBy"]["__typename"] == "User":
        return wdt.User.from_gql(artifactVersion.gql["createdBy"])
    return None


# Section 5/6: Connection Ops
gql_connection_op(
    "artifactVersion-artifactCollections",
    wdt.ArtifactVersionType,
    "artifactCollections",
    wdt.ArtifactCollectionType,
)

gql_connection_op(
    "artifactVersion-memberships",
    wdt.ArtifactVersionType,
    "artifactMemberships",
    wdt.ArtifactCollectionMembershipType,
)

gql_connection_op(
    "artifactVersion-usedBy",
    wdt.ArtifactVersionType,
    "usedBy",
    wdt.RunType,
)


# Section 6/6: Non Standard Business Logic Ops
@op(
    name="artifactVersion-name",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            versionIndex
            artifactSequence {
                id
                name
            }
        """,
    ),
)
def op_artifact_version_name(
    artifact: wdt.ArtifactVersion,
) -> str:
    return f'{artifact.gql["artifactSequence"]["name"]}:v{artifact.gql["versionIndex"]}'


@op(
    name="artifactVersion-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            versionIndex
            artifactSequence {
                id
                name
                defaultArtifactType {
                    id
                    name
                    project {
                        id
                        name
                        entity {
                            id
                            name
                        }
                    }
                }
            }
        """,
    ),
)
def artifact_version_link(
    artifactVersion: wdt.ArtifactVersion,
) -> wdt.Link:
    home_sequence_name = artifactVersion.gql["artifactSequence"]["name"]
    home_sequence_version_index = artifactVersion.gql["versionIndex"]
    type_name = artifactVersion.gql["artifactSequence"]["defaultArtifactType"]["name"]
    project_name = artifactVersion.gql["artifactSequence"]["defaultArtifactType"][
        "project"
    ]["name"]
    entity_name = artifactVersion.gql["artifactSequence"]["defaultArtifactType"][
        "project"
    ]["entity"]["name"]
    return wdt.Link(
        f"{home_sequence_name}:v{home_sequence_version_index}",
        f"/{entity_name}/{project_name}/artifacts/{quote(type_name)}/{quote(home_sequence_name)}/v{home_sequence_version_index}",
    )


# The following op: artifactVersion-isWeaveObject
# need more work to get artifact version file metadata.


@op(
    name="artifactVersion-isWeaveObject",
    plugins=wb_gql_op_plugin(lambda inputs, inner: "metadata"),
)
def artifact_version_is_weave_object(
    artifactVersion: wdt.ArtifactVersion,
) -> bool:
    raw_meta = artifactVersion.gql.get("metadata") or "{}"
    raw_meta_dict = json.loads(raw_meta)
    return raw_meta_dict.get("_weave_meta", {}).get("is_weave_obj", False)


@op(
    name="artifactVersion-weaveType",
    plugins=wb_gql_op_plugin(lambda inputs, inner: static_art_file_gql),
)
def artifact_version_weave_type(
    artifactVersion: wdt.ArtifactVersion,
) -> types.Type:
    return artifact_fs.FilesystemArtifactRef(
        _artifact_version_to_wb_artifact(artifactVersion), "obj"
    ).type


def _get_history_metrics(
    artifactVersion: wdt.ArtifactVersion,
) -> dict[str, typing.Any]:
    from ..compile import enable_compile
    from weave.graph import OutputNode, ConstNode
    from . import wb_domain_types
    from .. import weave_internal

    created_by = artifactVersion.gql["createdBy"]
    if created_by["__typename"] != "Run":
        return {}

    run_name = created_by["name"]
    project_name = created_by["project"]["name"]
    entity_name = created_by["project"]["entity"]["name"]
    history_step = artifactVersion.gql["historyStep"]

    node = OutputNode(
        types.TypedDict({}),
        "run-historyAsOf",
        {
            "run": OutputNode(
                wb_domain_types.RunType,
                "project-run",
                {
                    "project": OutputNode(
                        wb_domain_types.ProjectType,
                        "root-project",
                        {
                            "entity_name": ConstNode(types.String(), entity_name),
                            "project_name": ConstNode(types.String(), project_name),
                        },
                    ),
                    "run_name": ConstNode(types.String(), run_name),
                },
            ),
            "asOfStep": ConstNode(types.Int(), history_step),
        },
    )

    with enable_compile():
        res = weave_internal.use(node)
    if isinstance(res, list):
        res = res[0]

    return res


# This op contains a bunch of custom logic, punting for now
@op(
    hidden=True,
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            historyStep
            createdBy {
                __typename
                ... on Run {
                    id
                    name
                    project {
                        id
                        name 
                        entity {
                            id
                            name
                        }
                    }
                }
            }
        """,
    ),
)
def refine_history_metrics(
    artifactVersion: wdt.ArtifactVersion,
) -> types.Type:
    return wb_util.process_run_dict_type(_get_history_metrics(artifactVersion))


@op(
    name="artifactVersion-historyMetrics",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            historyStep
            createdBy {
                __typename
                ... on Run {
                    id
                    name
                    project {
                        id
                        name 
                        entity {
                            id
                            name
                        }
                    }
                }
            }
        """,
    ),
    refine_output_type=refine_history_metrics,
)
def history_metrics(
    artifactVersion: wdt.ArtifactVersion,
) -> dict[str, typing.Any]:
    return _get_history_metrics(artifactVersion)


# Special bridge functions to lower level local artifacts


# TODO: Move all this to helper functions off the artifactVersion object
def _artifact_version_to_wb_artifact(artifactVersion: wdt.ArtifactVersion):
    entity_name = artifactVersion.gql["artifactSequence"]["defaultArtifactType"][
        "project"
    ]["entity"]["name"]
    project_name = artifactVersion.gql["artifactSequence"]["defaultArtifactType"][
        "project"
    ]["name"]
    type_name = artifactVersion.gql["artifactSequence"]["defaultArtifactType"]["name"]
    home_sequence_name = artifactVersion.gql["artifactSequence"]["name"]
    commit_hash = artifactVersion.gql["commitHash"]
    return artifact_wandb.WandbArtifact(
        name=home_sequence_name,
        type=type_name,
        uri=artifact_wandb.WeaveWBArtifactURI(
            home_sequence_name, commit_hash, entity_name, project_name
        ),
    )


@op(
    name="artifactVersion-_file_refine_output_type",
    hidden=True,
    output_type=types.TypeType(),
    plugins=wb_gql_op_plugin(lambda inputs, inner: static_art_file_gql),
)
def _file_refine_output_type(artifactVersion: wdt.ArtifactVersion, path: str):
    art_local = _artifact_version_to_wb_artifact(artifactVersion)
    return types.TypeRegistry.type_of(art_local.path_info(path))


# Warning: the return type of this is incorrect! Weave0 treats
# type 'file' (FilesystemArtifactFile) as both dir and file.
# We have a refiner to do the correct thing, but the return
# type is set to `File` so that the first non-refine compile
# path will still work.
@op(
    name="artifactVersion-file",
    refine_output_type=_file_refine_output_type,
    plugins=wb_gql_op_plugin(lambda inputs, inner: static_art_file_gql),
)
def file_(
    artifactVersion: wdt.ArtifactVersion, path: str
) -> typing.Union[
    None, artifact_fs.FilesystemArtifactFile  # , artifact_fs.FilesystemArtifactDir
]:
    art_local = _artifact_version_to_wb_artifact(artifactVersion)
    return art_local.path_info(path)  # type: ignore


@op(
    name="artifactVersion-files",
    plugins=wb_gql_op_plugin(lambda inputs, inner: static_art_file_gql),
)
def files(
    artifactVersion: wdt.ArtifactVersion,
) -> list[artifact_fs.FilesystemArtifactFile]:
    art_local = _artifact_version_to_wb_artifact(artifactVersion)
    return artifact_wandb.FilesystemArtifactFileIterator(art_local)
